# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# vim: ts=4 sw=4 sts=0 noexpandtab:
import os
import time
from functools import wraps
from io import BytesIO
from pathlib import Path
import numpy

from ..logging import logger


def expand_input_format_compatibility(func):
    """
    Decorator to enhance the compatibility of a function that expects a file path
    string or a file-like object as its first argument.

    If the first argument is provided as bytes, it is wrapped in a BytesIO. If it
    is provided as a pathlib.Path instance, it is converted to a string representation
    of the path.

    Parameters
    ----------
    func : callable
        The function to be decorated. It is expected that the function's first argument
        is a file path string or a file-like object.

    Returns
    -------
    callable
        The decorated function with enhanced input compatibility.

    Example
    -------
    @expand_input_format_compatibility
    def read_data(file):
        pass

    read_data(b"some_bytes_data")  # Bytes are wrapped into BytesIO
    read_data(Path("/path/to/file"))  # Path is converted to string

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Assuming the first argument to the function is the input in question
        input_data = args[0]

        # Check if input is bytes, if yes, wrap it in BytesIO
        if isinstance(input_data, bytes):
            args = (BytesIO(input_data),) + args[1:]

        # Check if input is a Path instance, if yes, convert to string
        elif isinstance(input_data, Path):
            args = (str(input_data),) + args[1:]

        return func(*args, **kwargs)

    return wrapper


def compress_file(func):
    """
    Decorator used to write compressed file to disk using gz or bz2 protocols
    """

    def wrapped_func(filename, *args, **kwargs):
        if not isinstance(filename):
            return func(filename, *args, **kwargs)

        if filename.endswith('gz'):
            import gzip
            f = gzip.open(filename, 'w')
        elif filename.endswith('bz2'):
            import bz2
            f = bz2.BZ2File(filename, 'w')
        elif filename.endswith('zip'):
            print('Zip protocol is not supported')

        else:
            f = filename

    return func(f, *args, **kwargs)


class buggy(object):
    """
    This is a decorator which can be used to mark functions
    as not implemented. It will result in a warning being emitted when
    the function is called.
    """

    def __init__(self, date):
        self.date = date

    def __call__(self, fct):
        @wraps(fct)
        def wrapper(*args, **kwargs):
            logger.warning(
                "%s is buggy, use at your own risk. Expected fix : %s" % (
                fct.__name__, self.date))
            try:
                return_value = fct(*args, **kwargs)
            except Exception:
                print(
                    "%s Crashed : Didn't i told you it was buggy ?" % fct.__name__)
                raise

            return return_value

        return wrapper


class unimplemented(object):
    """
    This is a decorator which can be used to mark functions
    as not implemented. It will result in a warning being emitted and
    the function is never called.
    """

    def __init__(self, date):
        self.date = date

    def __call__(self, fct):
        @wraps(fct)
        def wrapper(*args, **kwargs):
            logger.warning("%s is not_implemented ... now. Expected : %s " % (
            fct.__name__, self.date))

        return wrapper


class broken(object):
    """
    This is a decorator which can be used to mark functions
    as broken. It will result in a warning being emitted and
    the function is never called.
    """

    def __init__(self, date):
        self.date = date

    def __call__(self, fct):
        @wraps(fct)
        def wrapper(*args, **kwargs):
            logger.critical(
                "%s is broken -- DON'T USE IT. Expected fix : %s" % (
                fct.__name__, self.date))

        return wrapper


def deprecated(fct):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    @wraps(fct)
    def wrapper(*args, **kwargs):
        logger.warning("Call to deprecated function %s." % fct.__name__)

        return fct(*args, **kwargs)

    return wrapper


def loggedcall(fct):
    """
    This is a decorator which log function call.
    """

    @wraps(fct)
    def wrapper(*args, **kwargs):
        logger.info(
            "Function -- %s--  called with arguments -- %s -- and keywords -- %s --" %
            (fct.__name__, str(args), str(kwargs)))
        return_value = fct(*args, **kwargs)
        logger.info("Function -- %s -- returned -- %s --" %
                    (fct.__name__, return_value))

        return return_value

    return wrapper


def addmethod(instance):
    def decorator(fct):
        setattr(instance, fct.func_name, fct)

        return fct

    return decorator


def timedcall(fct):
    @wraps(fct)
    def wrapper(*args, **kwargs):
        t = time.time()
        return_value = fct(*args, **kwargs)
        logger.info("Function -- %s -- called : TIME -- %.4f --" %
                    (fct.__name__, time.time() - t))

        return return_value

    return wrapper


class memoizehd(object):
    """
    This is a decorator which is designed to cache to a file the
    result of a function. This function calculate a hash from the
    function and the parameters and store the result in a designed
    file. ONLY WORK WITH SCIPY/NUMPY ARRAY AND WITH HASHABLE PARAMETERS
    """

    def __init__(self, basepath='/tmp'):
        basepath = basepath[1:] if basepath.startswith('/') else basepath
        basepath = os.path.join("/tmp", basepath)
        try:
            os.stat(basepath)
        except:
            os.mkdir(basepath)
        self.basepath = basepath
        print(self.basepath)

    def __call__(self, fct):
        cachebase = os.path.join(self.basepath,
                                 str(hash(fct)) + "_" + str(os.getpid()))

        @wraps(fct)
        def wrapper(*args, **kwargs):
            cachefile = cachebase + ''.join([str(i) for i in map(hash, args)])
            cachefile += '_'.join(
                [str(hash(kwargs[i])) for i in kwargs]) + ".npy"
            try:
                os.stat(cachefile)
                logger.info(
                    "Parameters hash matched calling -- %s --, reading cached return from file " %
                    fct.__name__)
                return_value = numpy.load(cachefile)
            except:
                return_value = fct(*args, **kwargs)
                numpy.save(cachefile, return_value)

            return return_value

        return wrapper


def memoize(fct):
    """
    This is a decorator which cache the result of a function based on the
    given parameter.
    """
    return_dict = {}

    @wraps(fct)
    def wrapper(*args, **kwargs):
        if args not in return_dict:
            return_dict[args] = fct(*args, **kwargs)

        return return_dict[args]

    return wrapper


def update_doc(cls):
    """
    Decorator function to update the docstring of a class.
    Inherits the documentation from its parent class, and allows
    for additional customization via class attributes:
    __remove_parameters__, __new_section__, and __example__.

    :param cls: The class whose docstring needs to be updated
    :return: The class with the updated docstring
    """

    # Inherit parent class doc if available (assuming single inheritance)
    parent_doc = cls.__bases__[0].__doc__ if cls.__bases__ else ''

    # Replace specific substrings in the inherited doc
    parent_doc = parent_doc.replace('~obspy.', '~uquake.')
    parent_doc = parent_doc.replace('obspy', 'uquake - powered by obspy -')

    # Append parent doc to current class doc
    if parent_doc:
        cls.__doc__ = f"{parent_doc}"

    original_doc = cls.__doc__

    # Remove specified parameters from the doc
    if hasattr(cls, '__remove_parameters__'):
        for section in cls.__remove_parameters__:
            start_idx = original_doc.find(f':type {section}')
            end_idx = original_doc.find(':type',
                                        original_doc.find(f':param {section}') + 1)

            # If section exists, remove it
            if start_idx != -1:
                original_doc = original_doc[:start_idx] + original_doc[end_idx:]

        cls.__doc__ = original_doc

    # Insert new doc sections if defined
    if hasattr(cls, '__new_section__'):
        new_section = cls.__new_section__['doc']
        previous_parameter = cls.__new_section__['previous_parameter']

        if previous_parameter:
            start_idx = cls.__doc__.find(f':type {previous_parameter}')
            end_idx = cls.__doc__.find(':type',
                                       cls.__doc__.find(f':param {previous_parameter}'))
            cls.__doc__ = cls.__doc__[:start_idx] + f'{new_section}' + cls.__doc__[
                                                                       end_idx:]
        else:
            cls.__doc__ += f'{new_section}'

    # Replace example section if defined
    if hasattr(cls, '__example__'):
        start_idx = original_doc.find('.. rubric:: Example')
        if start_idx != -1:
            end_idx = original_doc.find('.. note::', start_idx)
            if end_idx == -1:
                end_idx = len(original_doc)

            original_doc = original_doc[:start_idx] + '.. rubric:: Example\n' + \
                           cls.__example__ + original_doc[end_idx:] + '\n'

        cls.__doc__ = original_doc

    return cls





from pathlib import Path
import numpy as np
from uquake.core.logging import logger


def _write_grid_data(grid, base_name, path='.'):

    Path(path).mkdir(parents=True, exist_ok=True)

    with open(Path(path) / (base_name + '.buf'), 'wb') \
            as out_file:
        if grid.float_type == 'FLOAT':
            out_file.write(grid.data.astype(np.float32).tobytes())

        elif grid.float_type == 'DOUBLE':
            out_file.write(grid.data.astype(np.float64).tobytes())


def _write_grid_header(grid, base_name, path='.', seed_label=None,
                       seed=None, seed_units=None):

    # convert 'METER' to 'KILOMETER'
    if grid.grid_units == 'METER':
        origin = grid.origin / 1000
        spacing = grid.spacing / 1000
    else:
        origin = grid.origin
        spacing = grid.spacing

    line1 = f'{grid.shape[0]:d} {grid.shape[1]:d} {grid.shape[2]:d}  ' \
            f'{origin[0]:f} {origin[1]:f} {origin[2]:f}  ' \
            f'{spacing[0]:f} {spacing[1]:f} {spacing[2]:f}  ' \
            f'{grid.grid_type}\n'

    with open(Path(path) / (base_name + '.hdr'), 'w') as out_file:
        out_file.write(line1)

        if grid.grid_type in ['TIME', 'ANGLE']:

            if seed_units is None:
                logger.warning(f'seed_units are not defined. '
                               f'Assuming same units as grid ('
                               f'{grid.grid_units}')
            if grid.grid_units == 'METER':
                seed = seed / 1000

            line2 = u"%s %f %f %f\n" % (seed_label,
                                        seed[0], seed[1], seed[2])
            out_file.write(line2)

        out_file.write(u'TRANSFORM  NONE\n')


def _write_grid_model_id(base_name, model_id, path='.'):
    with open(Path(path) / (base_name + '.mid'), 'w') as out_file:
        out_file.write(f'{model_id}')


def write_nlloc(base_name, path='.'):

    _write_grid_data(base_name, path=path)
    _write_grid_header(base_name, path=path)
    _write_grid_model_id(base_name, path=path)



import warnings
from obspy.core.event import ResourceIdentifier
import numpy as np
import base64
import io
from obspy.core.util import AttribDict
from uquake.core.event import obsevent


def array_to_b64(array):
    output = io.BytesIO()
    np.save(output, array)
    content = output.getvalue()
    encoded = base64.b64encode(content).decode('utf-8')
    return encoded


def isfloat(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def b64_to_array(b64str):
    arr = np.load(io.BytesIO(base64.b64decode(b64str)))
    return arr


def pop_keys_matching(dict_in, keys):
    # Move keys from dict_in to dict_out
    dict_out = {}
    for key in keys:
        if key in dict_in:
            dict_out[key] = dict_in.pop(key)
    return dict_out


def parse_string_val(val, arr_flag='npy64_'):
    """
    Parse extra args in quakeML which are all stored as string.
    """
    if val is None:  # hack for deepcopy ignoring isfloat try-except
        val = None
    elif type(val) == AttribDict:
        val = val
    elif isfloat(val):
        val = float(val)
    elif str(val) == 'None':
        val = None
    elif val[:len(arr_flag)] == 'npy64_':
        val = b64_to_array(val[len(arr_flag):])
    return val


class UtilityMixin:
    def _set_attr_handler(self, name, value,
                          namespace='https://microquake.ai/xml/event/1'):
        """
        Generic handler to set attributes for uquake objects
        which inherit from ObsPy objects. If 'name' is not in
        default keys then it will be set in self['extra'] dict. If
        'name' is not in default keys but in the self.extra_keys
        then it will also be set as a class attribute. When loading
        extra keys from quakeml file, those in self.extra_keys will
        be set as attributes.
        """

        #  use obspy default setattr for default keys
        if name in self.defaults.keys():
            self.__dict__[name] = value
            if isinstance(self.__dict__[name], ResourceIdentifier):
                self.__dict__[name] = ResourceIdentifier(id=value.id)
            # super(type(self), self).__setattr__(name, value)
        elif name in self.extra_keys:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self[name] = value
            if hasattr(value, 'to_json'):
                value = value.to_json()
            elif isinstance(value, np.ndarray):
                value = "npy64_" + array_to_b64(value)
            elif type(value) is str:
                if "npy64_" in value:
                    value.replace("npy64_", "")
                    b64_to_array(value)
            self['extra'][name] = {'value': value, 'namespace': namespace}
        # recursive parse of 'extra' args when constructing uquake from obspy
        elif name == 'extra':
            if 'extra' not in self:  # hack for deepcopy to work
                self['extra'] = {}
            for key, adict in value.items():
                if key in self.extra_keys:
                    self.__dict__[key] = parse_string_val(adict.value)
                else:
                    self['extra'][key] = adict
        else:
            raise KeyError(name)

    def _init_handler(self, obspy_obj, **kwargs):
        """
        Handler to initialize uquake objects which
        inherit from ObsPy class. If obspy_obj is none,
        Kwargs is expected to be a mix of obspy kwargs
        and uquake kwargs specified by the hardcoded
        extra_keys.
        """

        if obspy_obj and len(kwargs) > 0:
            raise AttributeError("Initialize from either \
                                  obspy_obj or kwargs, not both")

        # default initialize the extra_keys args to None
        self['extra'] = {}
        [self.__setattr__(key, None) for key in self.extra_keys]

        if obspy_obj:
            self._init_from_obspy_object(self, obspy_obj)

            if 'resource_id' in obspy_obj.__dict__.keys():
                rid = obspy_obj.resource_id.id
                self.resource_id = ResourceIdentifier(id=rid,
                                                      referred_object=self)
        else:
            extra_kwargs = pop_keys_matching(kwargs, self.extra_keys)
            super(type(self), self).__init__(**kwargs)  # init obspy_origin args
            [self.__setattr__(k, v) for k, v in extra_kwargs.items()]  # init
            # extra_args

    def _init_from_obspy_object(self, uquake_obj, obspy_obj):
        """
        When initializing uquake object from obspy_obj
        checks attributes for lists of obspy objects and
        converts them to equivalent uquake objects.
        """

        class_equiv = {obsevent.event: Event,
                       obsevent.Pick: Pick,
                       obsevent.Arrival: Arrival,
                       obsevent.Origin: Origin,
                       obsevent.Magnitude: Magnitude,
                       obsevent.WaveformStreamID: WaveformStreamID}

        for key, val in obspy_obj.__dict__.items():
            itype = type(val)
            if itype in class_equiv:
                uquake_obj.__setattr__(key, class_equiv[itype](val))

            elif itype == list:
                out = []
                for item in val:
                    itype = type(item)
                    if itype in class_equiv:
                        out.append(class_equiv[itype](item))
                    else:
                        out.append(item)
                uquake_obj.__setattr__(key, out)
            else:
                uquake_obj.__setattr__(key, val)






# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: inventory.py
#  Purpose: Expansion of the obspy.core.inventory.inventory module
#   Author: uquake development team
#    Email: devs@uquake.org
#
# Copyright (C) 2016 uquake development team
# --------------------------------------------------------------------
"""
Expansion of the obspy.core.event module

:copyright:
    uquake development team (devs@uquake.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
import io
from obspy.core import inventory, AttribDict, UTCDateTime
from obspy.core.inventory import Network
from obspy.signal.invsim import corn_freq_2_paz
from obspy.core.inventory import (Response, InstrumentSensitivity,
                                  PolesZerosResponseStage, ResponseStage)
import numpy as np
from obspy.core.inventory.util import (Equipment, Operator, Person,
                                       PhoneNumber, Site, _textwrap,
                                       _unified_content_strings)
from uquake.core.util.decorators import expand_input_format_compatibility
from uquake.core.coordinates import Coordinates
from pathlib import Path

from .logging import logger
from uquake import __package_name__ as ns

import pandas as pd
from io import BytesIO
from .util.tools import lon_lat_x_y

from .util import ENTRY_POINTS
from pkg_resources import load_entry_point
from tempfile import NamedTemporaryFile
import os
from .util.requests import download_file_from_url
from uquake.core.util.attribute_handler import set_extra, get_extra, namespace


class SystemResponse(object):
    def __init__(self, sensor_type, **sensor_params):
        def __init__(self, sensor_type, **sensor_params):
            """
            Initialize a SystemResponse object.

            Parameters:
            -----------
            sensor_type : str
                Type of the sensor. Accepted values are 'geophone' or 'accelerometer'.

            **sensor_params : keyword arguments
                Additional parameters for configuring the sensor response. These vary based on the `sensor_type`.

                For 'geophone':
                - resonance_frequency : float
                    The resonance frequency of the geophone, in Hz.
                - gain : float
                    The gain of the geophone response.
                - damping : float, optional
                    Damping factor. Default is 0.707.
                - stage_sequence_number : int, optional
                    Stage sequence number. Default is 1.

                For 'accelerometer':
                - resonance_frequency : float
                    The resonance frequency of the accelerometer, in Hz.
                - gain : float
                    The gain of the accelerometer response.
                - sensitivity : float, optional
                    Sensitivity of the accelerometer. Default is 1.
                - damping : float, optional
                    Damping factor. Default is 0.707.
                - stage_sequence_number : int, optional
                    Stage sequence number. Default is 1.

            Raises:
            --------
            ValueError
                When an invalid `sensor_type` or unsupported `sensor_params` are supplied.

            Examples:
            ----------
            >>> sr = SystemResponse('geophone', resonance_frequency=4.5, gain=2.0)
            >>> sr = SystemResponse('accelerometer', resonance_frequency=15, gain=3.0, sensitivity=0.2)
            """

        self.components_info = {
            'sensor': {'type': sensor_type,
                       'params': sensor_params}
        }

    def add_cable(self, **cable_params):
        if 'cable' in self.components_info:
            raise ValueError("Cable response already added.")
        self.components_info['cable'] = {'params': cable_params}

    def add_digitizer(self, **digitizer_params):
        if 'digitizer' in self.components_info:
            raise ValueError("Digitizer already added.")
        self.components_info['digitizer'] = {'params': digitizer_params}

    def validate(self):
        if 'sensor' not in self.components_info:
            raise ValueError("System must have at least a sensor.")

    def generate_stage(self, component_key, stage_sequence_number):
        component_info = self.components_info.get(component_key, {})
        if component_key == 'sensor':
            if component_info['type'] == 'geophone':
                stage = geophone_sensor_response(**component_info['params'])
            elif component_info['type'] == 'accelerometer':
                stage = accelerometer_sensor_response(**component_info['params'])
            else:
                raise ValueError("Invalid sensor type.")
        elif component_key == 'cable':
            stage = sensor_cable_response(**component_info['params'])
        # Add more conditions for other components
        stage.stage_sequence_number = stage_sequence_number
        return stage

    @property
    def response(self):
        self.validate()
        response_stages = []
        sequence_number = 1
        for key in ['sensor', 'cable', 'digitizer']:
            if key in self.components_info:
                stage = self.generate_stage(key, sequence_number)
                response_stages.append(stage)
                sequence_number += 1

        # Create and return a Response object using the gathered response stages
        system_response = Response(response_stages=response_stages)

        return system_response


def geophone_sensor_response(resonance_frequency, gain, damping=0.707,
                             stage_sequence_number=1):
    """
    Generate a Poles and Zeros response stage for a geophone.

    Parameters:
    - resonance_frequency (float): Resonance frequency of the geophone in Hz.
    - gain (float): Gain factor.
    - damping (float, optional): Damping ratio. Defaults to 0.707 (critical damping).
    - stage_sequence_number (int, optional): Sequence number for the response stage.

    Returns:
    - PolesZerosResponseStage: Response stage object with configured poles and zeros.

    Notes:
    - The function utilizes the Laplace transform in the frequency domain.
    """
    paz = corn_freq_2_paz(resonance_frequency, damp=damping)
    pzr = PolesZerosResponseStage(stage_sequence_number, gain,
                                  resonance_frequency, 'M/S', 'V',
                                  'LAPLACE (RADIANT/SECOND)',
                                  resonance_frequency, paz['zeros'],
                                  paz['poles'])
    return pzr


def accelerometer_sensor_response(resonance_frequency, gain, sensitivity=1,
                                  damping=0.707, stage_sequence_number=1):
    """
    Generate a Poles and Zeros response stage for an accelerometer.

    Parameters:
    - resonance_frequency (float): Resonance frequency of the accelerometer in Hz.
    - gain (float): Gain factor.
    - sensitivity (float, optional): Sensitivity of the sensor. Defaults to 1.
    - damping (float, optional): Damping ratio. Defaults to 0.707 (critical damping).
    - stage_sequence_number (int, optional): Sequence number for the response stage.

    Returns:
    - PolesZerosResponseStage: Response stage object with configured poles and zeros.

    Notes:
    - The function utilizes the Laplace transform in the frequency domain.
    """
    paz = corn_freq_2_paz(resonance_frequency, damp=damping)
    paz['zeros'] = []
    pzr = PolesZerosResponseStage(stage_sequence_number, gain,
                                  resonance_frequency, 'M/S/S', 'V',
                                  'LAPLACE (RADIANT/SECOND)',
                                  resonance_frequency, paz['zeros'],
                                  paz['poles'])
    return pzr


def sensor_cable_response(output_resistance=np.inf, cable_length=np.inf,
                          cable_capacitance=np.inf, stage_sequence_number=2):
    """
    Generate a Poles and Zeros response stage for sensor-cable interaction.

    Parameters:
    - output_resistance (float, optional): Output resistance of the sensor in ohms. D
    efaults to infinity.
    - cable_length (float, optional): Length of the cable in meters. Defaults to
    infinity.
    - cable_capacitance (float, optional): Cable capacitance in farads per meter.
    Defaults to infinity.
    - stage_sequence_number (int, optional): Sequence number for the response stage.

    Returns:
    - PolesZerosResponseStage: Response stage object with configured poles.

    Notes:
    - Assumes a Laplace transform in the frequency domain for the response.
    """
    R = output_resistance
    l = cable_length
    C = cable_capacitance
    poles = []

    if ((R * l * C) != np.inf) and ((R * l * C) != 0):
        pole_cable = -1 / (R * l * C)
        poles.append(pole_cable)

    pzr = PolesZerosResponseStage(stage_sequence_number, 1,
                                  0, 'V', 'V',
                                  'LAPLACE (RADIANT/SECOND)',
                                  0, [], poles)
    return pzr


def get_response_from_nrl(datalogger_keys, sensor_keys):
    pass


class Inventory(inventory.Inventory):

    # def __init__(self, *args, **kwargs):
    #     super().__init__(self, *args, **kwargs)

    @classmethod
    def from_obspy_inventory_object(cls, obspy_inventory,
                                    xy_from_lat_lon=False,
                                    input_projection=4326,
                                    output_projection=None):

        source = ns  # Network ID of the institution sending
        # the message.

        inv = cls([], ns)
        inv.networks = []
        for network in obspy_inventory.networks:
            inv.networks.append(Network.from_obspy_network(network,
                                                           xy_from_lat_lon=
                                                           xy_from_lat_lon,
                                                           input_projection=
                                                           input_projection,
                                                           output_projection=
                                                           output_projection))

        return inv

    @staticmethod
    def from_url(url):
        """
        Load an ObsPy inventory object from a URL.

        :param url: The URL to download the inventory file from.
        :type url: str
        :return: The loaded ObsPy inventory object.
        :rtype: obspy.core.inventory.Inventory
        """
        # Download the inventory file from the URL

        inventory_data = download_file_from_url(url)

        # Save the inventory data to a temporary file
        with NamedTemporaryFile(delete=False) as temp_file:
            temp_file.write(inventory_data.read())
            temp_file.flush()

            # Load the inventory from the temporary file
            file_format = 'STATIONXML'  # Replace with the correct format
            inventory = read_inventory(temp_file.name, format=file_format)

        # Remove the temporary file after reading the inventory
        os.remove(temp_file.name)

        return inventory

    @staticmethod
    def from_bytes(byte_string):
        file_in = io.BytesIO(byte_string)
        file_in.read()
        file_in.seek(0)
        return read_inventory(file_in)

    def write(self, path_or_file_obj, *args, format='stationxml',
              nsmap={'mq': namespace}, **kwargs):
        return super().write(path_or_file_obj, *args, format=format,
                             nsmap=nsmap, **kwargs)

    def get_station(self, sta):
        return self.select(sta)

    def get_channel(self, sta=None, cha=None):
        return self.select(sta, cha_code=cha)

    def select_instrument(self, instruments=None):
        if isinstance(instruments, list):
            for location in instruments:
                for obj_site in self.instruments:
                    if location.code == obj_site.code:
                        yield obj_site

        elif isinstance(instruments, str):
            location = instruments
            for obj_site in self.instruments:
                if location.code == obj_site.code:
                    return obj_site

    def to_bytes(self):

        file_out = BytesIO()
        self.write(file_out)
        return file_out.getvalue()

    def __eq__(self, other):
        return np.all(self.instruments == other.instruments)

    # def write(self, filename):
    #     super().write(self, filename, format='stationxml', nsmap={ns: ns})

    @property
    def instruments(self):
        instruments = []
        for network in self.networks:
            for station in network.stations:
                for instrument in station.instruments:
                    instruments.append(instrument)

        return np.sort(instruments)


class Network(inventory.Network):
    __doc__ = inventory.Network.__doc__.replace('obspy', ns)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_obspy_network(cls, obspy_network, xy_from_lat_lon=False,
                           input_projection=4326, output_projection=None):

        net = cls(obspy_network.code)

        for key in obspy_network.__dict__.keys():
            if 'stations' in key:
                net.__dict__[key] = []
            else:
                try:
                    net.__dict__[key] = obspy_network.__dict__[key]
                except Exception as e:
                    logger.error(e)

        for i, station in enumerate(obspy_network.stations):
            net.stations.append(Station.from_obspy_station(station,
                                                           xy_from_lat_lon,
                                                           input_projection=
                                                           input_projection,
                                                           output_projection=
                                                           output_projection))

        return net

    def get_grid_extent(self, padding_fraction=0.1, ignore_stations=[],
                        ignore_sites=[]):
        """
        return the extents of a grid encompassing all sensors comprised in the
        network
        :param padding_fraction: buffer to add around the stations
        :param ignore_stations: list of stations to exclude from the
        calculation of the grid extents
        :param ignore_sites: list of location to exclude from the calculation of
        the grid extents
        :type ignore_stations: list
        :return: return the lower and upper corners
        :rtype: dict
        """

        xs = []
        ys = []
        zs = []

        coordinates = []
        for station in self.stations:
            if station.code in ignore_stations:
                continue
            for location in station.instruments:
                if location.code in ignore_sites:
                    continue
                coordinates.append(location.loc)

        min = np.min(coordinates, axis=0)
        max = np.max(coordinates, axis=0)
        # center = min + (max - min) / 2
        center = (min + max) / 2
        d = (max - min) * (1 + padding_fraction)

        c1 = center - d / 2
        c2 = center + d / 2

        return c1, c2

    @property
    def site_coordinates(self):
        coordinates = []
        for station in self.stations:
            coordinates.append(station.site_coordinates)

        return np.array(coordinates)

    @property
    def station_coordinates(self):
        return np.array([station.loc for station in self.stations])

    @property
    def instruments(self):
        instruments = []
        for station in self.stations:
            for instrument in station.instruments:
                instruments.append(instrument)
        return instruments

    @property
    def instrument_coordinates(self):
        coordinates = []
        for instrument in self.instruments:
            coordinates.append(instrument.loc)


class Station(inventory.Station):
    __doc__ = inventory.Station.__doc__.replace('obspy', ns)

    def __init__(self, *args, coordinates: Coordinates = Coordinates(0, 0, 0), **kwargs):

        if 'extra' not in self.__dict__.keys():  # hack for deepcopy to work
            self['extra'] = {}

        self.extra = AttribDict()

        if 'latitude' not in kwargs.keys():
            kwargs['latitude'] = 0
        if 'longitude' not in kwargs.keys():
            kwargs['longitude'] = 0
        if 'elevation' not in kwargs.keys():
            kwargs['elevation'] = 0

        # initialize the extra key

        if not hasattr(self, 'extra'):
            self.extra = AttribDict()

        super().__init__(*args, **kwargs)

        self.extra['coordinates'] = coordinates.to_extra_key(namespace=namespace)

    def __setattr__(self, name, value):
        if name == 'coordinates':
            self.extra[name] = value.to_extra_key(namespace=namespace)
        else:
            super().__setattr__(name, value)

    def __getattr__(self, item):
        if item == 'coordinates':
            return Coordinates.from_extra_key(self.extra[item])
        else:
            super().__getattr__(item)

    # def __getitem__(self, key):
    #     if isinstance(key, int):
    #         return self[key]
    #     else:
    #         return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__['key'] = value

    def __eq__(self, other):
        if not super().__eq__(other):
            return False

        return self.coordinates == other.coordinates

    @classmethod
    def from_obspy_station(cls, obspy_station, xy_from_lat_lon=False,
                           output_projection=None,
                           input_projection=4326):

        #     cls(*params) is same as calling Station(*params):

        stn = cls(obspy_station.code, latitude=obspy_station.latitude,
                  longitude=obspy_station.longitude,
                  elevation=obspy_station.elevation)
        for key in obspy_station.__dict__.keys():
            try:
                stn.__dict__[key] = obspy_station.__dict__[key]
            except Exception as e:
                logger.error(e)

        if xy_from_lat_lon:
            if (stn.latitude is not None) and (stn.longitude is not None):

                stn.x, stn.y = lon_lat_x_y(input_projection, output_projection,
                                           longitude=stn.longitude,
                                           latitude=stn.latitude)

                stn.z = obspy_station.elevation

            else:
                logger.warning(f'Latitude or Longitude are not'
                               f'defined for station {obspy_station.code}.')

                output_projection = 32725

        stn.channels = []

        for channel in obspy_station.channels:
            stn.channels.append(Channel.from_obspy_channel(channel,
                                                           xy_from_lat_lon=
                                                           xy_from_lat_lon,
                                                           input_projection=
                                                           input_projection,
                                                           output_projection=
                                                           output_projection))

        return stn

    @property
    def x(self):
        return self.coordinates.x

    @property
    def y(self):
        return self.coordinates.y

    @property
    def z(self):
        return self.coordinates.z

    @property
    def loc(self):
        if self.extra:
            if self.extra.get('x', None) and self.extra.get(
                    'y', None) and self.extra.get('z', None):
                return np.array([self.x, self.y, self.z])
            else:
                raise AttributeError
        else:
            raise AttributeError

    @property
    def instruments(self):
        location_codes = []
        channel_dict = {}
        instruments = []
        for channel in self.channels:
            location_codes.append(channel.location_code)
            channel_dict[channel.location_code] = []

        for channel in self.channels:
            channel_dict[channel.location_code].append(channel)

        for key in channel_dict.keys():
            instruments.append(Instrument(self, channel_dict[key]))

        return instruments

    @property
    def instrument_coordinates(self):
        coordinates = []
        for instrument in self.instruments:
            coordinates.append(instrument.loc)
        return np.array(coordinates)

    def __str__(self):
        contents = self.get_contents()

        x = self.x
        y = self.y
        z = self.z

        location_count = len(self.instruments)
        channel_count = len(self.channels)

        format_dict = {
            'code': self.code or 'N/A',
            'location_count': location_count or 0,
            'channel_count': channel_count or 0,
            'start_date': self.start_date or 'N/A',
            'end_date': self.end_date or 'N/A',
            'x': f"{x:.0f}" if x is not None else 0,
            'y': f"{y:.0f}" if y is not None else 0,
            'z': f"{z:.0f}" if z is not None else 0
        }

        ret = ("\tStation Code: {code}\n"
               "\tLocation Count: {location_count}\n"
               "\tChannel Count: {channel_count}\n"
               "\tDate range: {start_date} - {end_date}\n"
               "\tx: {x}, y: {y}, z: {z} m\n").format(**format_dict)

        if getattr(self, 'extra', None):
            if getattr(self.extra, 'x', None) and getattr(self.extra, 'y',
                                                          None):
                x = self.x
                y = self.y
                z = self.z
                ret = ("Station {station_name}\n"
                       "\tStation Code: {station_code}\n"
                       "\tLocation Count: {site_count}\n"
                       "\tChannel Count: {selected}/{total}"
                       " (Selected/Total)\n"
                       "\tDate range: {start_date} - {end_date}\n"
                       "\tEasting [x]: {x:.0f} m, Northing [y] m: {y:.0f}, "
                       "Elevation [z]: {z:.0f} m\n")

        ret = ret.format(
            station_name=contents["stations"][0],
            station_code=self.code,
            site_count=len(self.instruments),
            selected=self.selected_number_of_channels,
            total=self.total_number_of_channels,
            start_date=str(self.start_date),
            end_date=str(self.end_date) if self.end_date else "",
            restricted=self.restricted_status,
            alternate_code="Alternate Code: %s " % self.alternate_code if
            self.alternate_code else "",
            historical_code="Historical Code: %s " % self.historical_code if
            self.historical_code else "",
            x=x, y=y, z=z)
        ret += "\tAvailable Channels:\n"
        ret += "\n".join(_textwrap(
            ", ".join(_unified_content_strings(contents["channels"])),
            initial_indent="\t\t", subsequent_indent="\t\t",
            expand_tabs=False))

        return ret

    def __repr__(self):
        return self.__str__()


class Instrument:
    """
    This class is a container for grouping the channels into coherent entity
    that are Instruments. From the uquake package perspective a station is
    the physical location where data acquisition instruments are grouped.
    One or multiple instruments can be connected to a station.
    """

    def __init__(self, station, channels):

        location_codes = []
        for channel in channels:
            location_codes.append(channel.location_code)
            if len(np.unique(location_codes)) > 1:
                raise ValueError('the channels in the channel list should have a unique '
                                 'location code')

        self.location_code = location_codes[0]

        if len(np.unique(location_codes)) > 1:
            logger.error('the channels in the channel list should have a'
                         'unique location code')
            raise KeyError

        self.station = station
        self.channels = channels

    def __repr__(self):
        ret = f'\tLocation {self.location_code}\n' \
              f'\tx: {self.x:.0f} m, y: {self.y:.0f} m z: {self.z:0.0f} m\n' \
              f'\tChannel Count: {len(self.channels)}'

        return ret

    def __str__(self):
        return self.instrument_code

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __gt__(self, other):
        return str(self) > str(other)

    @property
    def loc(self):
        return np.array([self.x, self.y, self.z])

    @property
    def alternate_code(self):
        return self.channels[0].alternative_code

    @property
    def x(self):
        return self.channels[0].x

    @property
    def y(self):
        return self.channels[0].y

    @property
    def z(self):
        return self.channels[0].z

    @property
    def coordinates(self):
        return self.channels[0].coordinates

    @property
    def station_code(self):
        return self.station.code

    @property
    def instrument_code(self):
        return self.code

    @property
    def simplified_code(self):
        return f'{self.station_code}{self.location_code}'

    @property
    def code(self):
        return self.make_instrument_code(self.station_code, self.location_code)

    @property
    def sensor_type_code(self):
        return self.channels[0].code[0:-1]

    @staticmethod
    def make_instrument_code(station_code, location_code):
        return f'{station_code}.{location_code}'


class Channel(inventory.Channel):
    defaults = {}

    __doc__ = inventory.Channel.__doc__.replace('obspy', ns)

    def __init__(self, code, location_code, active: bool = True, oriented: bool = False,
                 coordinates: Coordinates = Coordinates(0, 0, 0),
                 orientation_vector=None, **kwargs):

        latitude = kwargs.pop('latitude') if 'latitude' in kwargs.keys() else 0
        longitude = kwargs.pop('longitude') if 'longitude' in kwargs.keys() else 0
        elevation = kwargs.pop('elevation') if 'elevation' in kwargs.keys() else 0
        depth = kwargs.pop('depth') if 'depth' in kwargs.keys() else 0

        super().__init__(code, location_code, latitude, longitude,
                                      elevation, depth, **kwargs)

        if 'extra' not in self.__dict__.keys():  # hack for deepcopy to work
            self.__dict__['extra'] = {}

        if orientation_vector is not None:
            # making the orientation vector (cosine vector) unitary
            orientation_vector = orientation_vector / np.linalg.norm(orientation_vector)
            self.set_orientation(orientation_vector)

        self.extra['coordinates'] = coordinates.to_extra_key(namespace=namespace)
        set_extra(self, 'active', active, namespace=namespace)
        set_extra(self, 'oriented', oriented, namespace=namespace)

    @classmethod
    def from_obspy_channel(cls, obspy_channel, xy_from_lat_lon=False,
                           output_projection=None, input_projection=4326):

        cha = cls(obspy_channel.code, obspy_channel.location_code,
                  latitude=obspy_channel.latitude,
                  longitude=obspy_channel.longitude,
                  elevation=obspy_channel.elevation,
                  depth=obspy_channel.depth)

        if hasattr(obspy_channel, 'extra'):
            for key in cha.extra.keys():
                if key not in obspy_channel.__dict__['extra'].keys():
                    cha.__dict__['extra'][key] = 0
                else:
                    cha.__dict__['extra'][key] = \
                        obspy_channel.__dict__['extra'][key]

        for key in obspy_channel.__dict__.keys():
            cha.__dict__[key] = obspy_channel.__dict__[key]

        if xy_from_lat_lon:
            if (cha.latitude is not None) and (cha.longitude is not None):

                cha.x, cha.y = lon_lat_x_y(input_projection, output_projection,
                                           longitude=cha.longitude,
                                           latitude=cha.latitude)

                cha.z = cha.elevation

        return cha

    def __getattr__(self, item):
        if item == 'coordinates':
            return Coordinates.from_extra_key(self.extra['coordinates'])
        elif item in ('active', 'oriented'):
            return get_extra(item)
        else:
            if hasattr(super(), item):
                return getattr(super(), item)
            else:
                raise AttributeError(
                    f"'{type(self).__name__}' object has no attribute '{item}'")

    def __setattr__(self, key, value):
        if key == 'coordinates':
            self.extra[key] = value.to_extra_key(namespace=namespace)
        elif key in ('active', 'oriented'):
            set_extra(self, key, value, namespace=namespace)
        else:
            super().__setattr__(key, value)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__['key'] = value

    def __repr__(self):
        time_range = f"{self.start_date} - {self.end_date}" if self.start_date and \
                                                               self.end_date else 'N/A'

        if self.coordinates.coordinate_system == 'ENU':

            attributes = {
                'Channel': self.code,
                'Location': self.location_code,
                'Time range': time_range,
                'Northing [x]': f"{self.x:0.0f} m" if self.x is not None else 'N/A',
                'Easting [y]': f"{self.y:0.0f} m" if self.y is not None else 'N/A',
                'Elevation [z]': f"{self.z:0.0f} m" if self.z is not None else 'N/A',
                'Dip (degrees)': f"{self.dip:0.0f}" if self.dip is not None else 'N/A',
                'Azimuth (degrees)': f"{self.azimuth:0.0f}" if self.azimuth
                                                               is not None else 'N/A',
                'Response information': 'available' if self.response else 'not available'
            }

        else:
            attributes = {
                'Channel': self.code,
                'Location': self.location_code,
                'Time range': time_range,
                'Easting [x]': f"{self.x:0.0f} m" if self.x is not None else 'N/A',
                'Northing [y]': f"{self.y:0.0f} m" if self.y is not None else 'N/A',
                'Depth [z]': f"{self.z:0.0f} m" if self.z is not None else 'N/A',
                'Dip (degrees)': f"{self.dip:0.0f}" if self.dip is not None else 'N/A',
                'Azimuth (degrees)': f"{self.azimuth:0.0f}" if self.azimuth
                                                               is not None else 'N/A',
                'Response information': 'available' if self.response else 'not available'
            }

        ret = "\n".join([f"{key}: {value}" for key, value in attributes.items()])
        return ret

    def __eq__(self, other):
        if not super().__eq__(other):
            return False

        return (self.coordinates == other.coordinates and
                self.active == other.active and
                self.oriented == other.oriented)

    def set_orientation(self, orientation_vector):
        """
        set the Azimuth and Dip from an orientation vector assuming the
        orientation vector provided is east, north, up.
        :param self:
        :param orientation_vector:
        :return:
        """

        east = orientation_vector[0]
        north = orientation_vector[1]
        up = orientation_vector[2]

        horizontal_length = np.linalg.norm([east, north])

        azimuth = np.arctan2(east, north) * 180 / np.pi
        if azimuth < 0:
            azimuth = 360 + azimuth

        self.azimuth = azimuth
        self.dip = np.arctan2(-up, horizontal_length) * 180 / np.pi

    @property
    def orientation_vector(self):

        up = -np.sin(self.dip * np.pi / 180)
        east = np.sin(self.azimuth * np.pi / 180) * \
               np.cos(self.dip * np.pi / 180)
        north = np.cos(self.azimuth * np.pi / 180) * \
                np.cos(self.dip * np.pi / 180)

        return np.array([east, north, up])

    @property
    def x(self):
        return self.coordinates.x

    @property
    def y(self):
        return self.coordinates.y

    @property
    def z(self):
        return self.coordinates.z

    @property
    def loc(self):
        return np.array([self.x, self.y, self.z])


def load_from_excel(file_name) -> Inventory:
    """
    Read in a multi-sheet excel file with network metadata sheets:
        Locations, Networks, Hubs, Stations, Components, Locations, Cables,
        Boreholes
    Organize these into a uquake Inventory object

    :param xls_file: path to excel file
    :type: xls_file: str
    :return: inventory
    :rtype: uquake.core.data.inventory.Inventory

    """

    df_dict = pd.read_excel(file_name, sheet_name=None)

    source = df_dict['Locations'].iloc[0]['code']
    # sender (str, optional) Name of the institution sending this message.
    sender = df_dict['Locations'].iloc[0]['operator']
    net_code = df_dict['Networks'].iloc[0]['code']
    net_descriptions = df_dict['Networks'].iloc[0]['name']

    contact_name = df_dict['Networks'].iloc[0]['contact_name']
    contact_email = df_dict['Networks'].iloc[0]['contact_email']
    contact_phone = df_dict['Networks'].iloc[0]['contact_phone']
    site_operator = df_dict['Locations'].iloc[0]['operator']
    site_country = df_dict['Locations'].iloc[0]['country']
    site_name = df_dict['Locations'].iloc[0]['name']
    location_code = df_dict['Locations'].iloc[0]['code']

    print("source=%s" % source)
    print("sender=%s" % sender)
    print("net_code=%s" % net_code)

    network = Network(net_code)
    inventory = Inventory([network], source)

    # obspy requirements for PhoneNumber are super specific:
    # So likely this will raise an error if/when someone changes the value in
    # Networks.contact_phone
    """
    PhoneNumber(self, area_code, phone_number, country_code=None,
    description=None):
        :type area_code: int
        :param area_code: The area code.
        :type phone_number: str
        :param phone_number: The phone number minus the country and
        area code. Must be in the form "[0-9]+-[0-9]+", e.g. 1234-5678.
        :type country_code: int, optional
        :param country_code: The country code.
    """

    import re
    phone = re.findall(r"[\d']+", contact_phone)
    area_code = int(phone[0])
    number = "%s-%s" % (phone[1], phone[2])
    phone_number = PhoneNumber(area_code=area_code, phone_number=number)

    person = Person(names=[contact_name], agencies=[site_operator],
                    emails=[contact_email], phones=[phone_number])
    operator = Operator(site_operator, contacts=[person])
    location = Instrument(name=site_name, description=site_name,
                          country=site_country)

    # Merge Stations+Components+Locations+Cables info into sorted stations +
    # channels dicts:

    df_dict['Stations']['station_code'] = df_dict['Stations']['code']
    df_dict['Locations']['sensor_code'] = df_dict['Locations']['code']
    df_dict['Components']['code_channel'] = df_dict['Components']['code']
    df_dict['Components']['sensor'] = df_dict['Components']['sensor__code']
    df_merge = pd.merge(df_dict['Stations'], df_dict['Locations'],
                        left_on='code', right_on='station__code',
                        how='inner', suffixes=('', '_channel'))

    df_merge2 = pd.merge(df_merge, df_dict['Components'],
                         left_on='sensor_code', right_on='sensor__code',
                         how='inner', suffixes=('', '_sensor'))

    df_merge3 = pd.merge(df_merge2, df_dict['Cable types'],
                         left_on='cable__code', right_on='code',
                         how='inner', suffixes=('', '_cable'))

    df_merge4 = pd.merge(df_merge3, df_dict['Location types'],
                         left_on='sensor_type__model', right_on='model',
                         how='inner', suffixes=('', '_sensor_type'))

    df = df_merge4.sort_values(['sensor_code', 'location_code']).fillna(0)

    # Need to sort by unique station codes, then look through 1-3 channels
    # to add
    stn_codes = set(df['sensor_code'])
    stations = []

    for code in stn_codes:
        chan_rows = df.loc[df['sensor_code'] == code]
        row = chan_rows.iloc[0]
        station = {}
        # Set some keys explicitly
        #     from ipdb import set_trace; set_trace()
        station['code'] = '{}'.format(row['sensor_code'])
        station['x'] = row['location_x_channel']
        station['y'] = row['location_y_channel']
        station['z'] = row['location_z_channel']
        station['loc'] = np.array(
            [station['x'], station['y'], station['z']])
        station['long_name'] = "{}.{}.{:02d}".format(row['network__code'],
                                                     row['station_code'],
                                                     row['location_code'])

        # MTH: 2019/07 Seem to have moved from pF to F on Cables sheet:
        station['cable_capacitance_pF_per_meter'] = row['c'] * 1e12

        # Set the rest (minus empty fields) directly from spreadsheet names:
        renamed_keys = {'sensor_code', 'location_x', 'location_y',
                        'location_z', 'name'}

        # These keys are either redundant or specific to channel, not station:
        remove_keys = {'code', 'id_channel', 'orientation_x',
                       'orientation_y', 'orientation_z', 'id_sensor',
                       'enabled_channel', 'station_id', 'id_cable'}
        keys = row.keys()
        empty_keys = keys[pd.isna(row)]
        keys = set(keys) - set(empty_keys) - renamed_keys - remove_keys

        for key in keys:
            station[key] = row[key]

        # Added keys:
        station['motion'] = 'VELOCITY'

        if row['sensor_type'].upper() == 'ACCELEROMETER':
            station['motion'] = 'ACCELERATION'

        # Attach channels:
        station['channels'] = []

        for index, rr in chan_rows.iterrows():
            chan = {}
            chan['cmp'] = rr['code_channel_sensor'].upper()
            chan['orientation'] = np.array([rr['orientation_x'],
                                            rr['orientation_y'],
                                            rr['orientation_z']])
            chan['x'] = row['location_x_channel']
            chan['y'] = row['location_y_channel']
            chan['z'] = row['location_z_channel']
            chan['enabled'] = rr['enabled']
            station['channels'].append(chan)

        stations.append(station)

    # from ipdb import set_trace; set_trace()

    # Convert these station dicts to inventory.Station objects and attach to
    # inventory.network:
    station_list = []

    for station in stations:
        # This is where namespace is first employed:
        station = Station.from_station_dict(station, site_name)
        station.location = location
        station.operators = [operator]
        station_list.append(station)

    network.stations = station_list

    return inventory


@expand_input_format_compatibility
def read_inventory(path_or_file_object, format='STATIONXML',
                   xy_from_lat_lon=False, input_projection=4326,
                   output_projection=None, *args, **kwargs) -> Inventory:
    """
    Read inventory file
    :param path_or_file_object: the path to the inventory file or a file object
    :param format: the format
    :param xy_from_lat_lon: if True convert populate the XY field by converting
    the latitude and longitude to UTM
    :param input_projection: The input projection. Applicable if
    xy_from_lat_lon is True (default=4326 for latitude longitude)
    :param output_projection: The output projection. Has to be specified if
    xy_from_lat_lon is True. Default=None
    :param args: see obspy.core.inventory.read_inventory for more information
    :param kwargs: see obspy.core.inventory.read_inventory for more information
    :return: an inventory object
    :rtype: ~uquake.core.inventory.Inventory
    """

    if xy_from_lat_lon and (output_projection is None):
        raise ValueError('the output projection is needed for conversion'
                         'from latitude-longitude to UTM')

    if type(path_or_file_object) is Path:
        path_or_file_object = str(path_or_file_object)

    # del kwargs['xy_from_lat_lon']

    if (format not in ENTRY_POINTS['inventory'].keys()) or \
            (format.upper() == 'STATIONXML'):

        obspy_inv = inventory.read_inventory(path_or_file_object, *args, format=format,
                                             **kwargs)

        return Inventory.from_obspy_inventory_object(
            obspy_inv, xy_from_lat_lon=xy_from_lat_lon,
            output_projection=output_projection,
            input_projection=input_projection)

    else:
        format_ep = ENTRY_POINTS['inventory'][format]

        read_format = load_entry_point(format_ep.dist.key,
                                       'obspy.io.%s' %
                                       format_ep.name, 'readFormat')

        return expand_input_format_compatibility(
            read_format(path_or_file_object, **kwargs))

    # else:


# def read_inventory(filename, format='STATIONXML', **kwargs):
#     if isinstance(filename, Path):
#         filename = str(filename)
#
#     if format in ENTRY_POINTS['inventory'].keys():
#         format_ep = ENTRY_POINTS['inventory'][format]
#         read_format = load_entry_point(format_ep.dist.key,
#                                        'obspy.plugin.inventory.%s' %
#                                        format_ep.name, 'readFormat')
#
#         return Inventory(inventory=read_format(filename, **kwargs))
#
#     else:
#         return inventory.read_inventory(filename, format=format,
#                                                **kwargs)


read_inventory.__doc__ = inventory.read_inventory.__doc__.replace(
    'obspy', ns)



# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: <filename>
#  Purpose: <purpose>
#   Author: <author>
#    Email: <email>
#
# Copyright (C) <copyright>
# --------------------------------------------------------------------
"""


:copyright:
    <copyright>
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
import random

import numpy as np
import scipy.ndimage

from .base import Grid
from pathlib import Path
from uuid import uuid4
import matplotlib.pyplot as plt
from loguru import logger
import skfmm
from multiprocessing import Pool, cpu_count
from functools import partial
from typing import Optional
import h5py
from .base import ray_tracer
import shutil
from uquake.grid import read_grid
# from .hdf5 import H5TTable, write_hdf5
from scipy.interpolate import interp1d
from uquake.core.event import WaveformStreamID
from enum import Enum
from typing import List
from uquake.core.coordinates import Coordinates, CoordinateSystem
from uquake.core.inventory import Inventory
from uquake.synthetic.inventory import generate_unique_instrument_code
from uquake.core.event import ResourceIdentifier
from .base import __default_grid_label__
import string

__cpu_count__ = cpu_count()

valid_phases = ('P', 'S')


class Phases(Enum):
    P = 'P'
    S = 'S'


class GridTypes(Enum):
    VELOCITY = 'VELOCITY'
    VELOCITY_METERS = 'VELOCITY_METERS'
    SLOWNESS = 'SLOWNESS'
    VEL2 = 'VEL2'
    SLOW2 = 'SLOW2'
    SLOW2_METERS = 'SLOW2_METERS'
    SLOW_LEN = 'SLOW_LEN'
    STACK = 'STACK'
    TIME = 'TIME'
    TIME2D = 'TIME2D'
    PROB_DENSITY = 'PROB_DENSITY'
    MISFIT = 'MISFIT'
    ANGLE = 'ANGLE'
    ANGLE2D = 'ANGLE2D'

    def __str__(self):
        return self.value


valid_grid_types = (
    'VELOCITY',
    'VELOCITY_METERS',
    'SLOWNESS',
    'VEL2',
    'SLOW2',
    'SLOW2_METERS',
    'SLOW_LEN',
    'STACK',
    'TIME',
    'TIME2D',
    'PROB_DENSITY',
    'MISFIT',
    'ANGLE',
    'ANGLE2D'
)


class FloatTypes(Enum):
    FLOAT = 'float32'
    DOUBLE = 'float64'

    def __str__(self):
        return self.value


valid_float_types = {
    # NLL_type: numpy_type
    'FLOAT': 'float32',
    'DOUBLE': 'float64'
}


class GridUnits(Enum):
    METER = 'METER'
    FEET = 'FEET'
    KILOMETER = 'KILOMETER'
    SECOND = 'SECOND'

    def __str__(self):
        return self.value


valid_grid_units = (
    'METER',
    'KILOMETER',
)

__velocity_grid_location__ = Path('model')
__time_grid_location__ = Path('time')


__default_grid_units__ = GridUnits.METER
__default_float_type__ = FloatTypes.FLOAT


class Seed:
    def __init__(self, station_code, location_code, coordinates: Coordinates,
                 elevation: float = 0):
        """
        Contains a location
        :param station_code: station code
        :param instrument_code: location code
        :param coordinates:
        """

        self.station = station_code
        self.location = location_code
        self.coordinates = coordinates
        self.elevation = elevation

    def __repr__(self):
        return f'label (station.location): {self.station}.{self.location}\n' \
               f'                       x: {self.coordinates.x:0.2f}\n' \
               f'                       y: {self.coordinates.y:0.2f}\n' \
               f'                       z: {self.coordinates.z:0.2f}\n' \
               f'               elevation: {self.elevation:0.2f}\n' \
               f'       coordinate_system: {self.coordinates.coordinate_system}\n'

    def __str__(self):
        return self.__repr__()

    @property
    def x(self):
        return self.coordinates.x

    @property
    def y(self):
        return self.coordinates.y

    @property
    def z(self):
        return self.coordinates.z

    @property
    def loc(self):
        return self.coordinates.loc

    @property
    def label(self):
        return f'{self.station}_{self.location}'

    @property
    def instrument_code(self):
        return self.label

    @property
    def T(self):
        return np.array(self.coordinates.loc).T

    @classmethod
    def random_in_grid(cls, grid: Grid, n_seeds: int = 1) -> 'List[Seed]':
        """
        Generate a random point within the grid boundary
        :param grid: a grid object
        :type grid: uquake.grid.base.Grid or an object inheriting from Grid
        :param n_seeds: number of locations to generate
        :type n_seeds: int
        :return: a list of Location objects

        :example:
        >> from uquake.grid.base import Grid
        >> from uquake.grid.nlloc import Location
        >> grid_dimensions = [10, 10, 10]
        >> grid_spacing = [1, 1, 1]
        >> grid_origin = [0, 0, 0]
        >> grid = Grid(grid_dimensions, grid_spacing, grid_origin, value=1)
        >> locations = Location.random_in_grid(grid, n_location=10)
        """

        codes = generate_unique_instrument_code(n_codes=n_seeds)
        seeds = []

        for i, (point, code) in enumerate(zip(grid.generate_random_points_in_grid(
                n_points=n_seeds), codes)):
            station_code = code.split('_')[0]
            location_code = code.split('_')[1]
            coordinates = Coordinates(point[0], point[1], point[2])
            seeds.append(cls(station_code, location_code, coordinates))

        return seeds


class SeedEnsemble:

    def __init__(self, seeds: List[Seed] = [], units: GridUnits = GridUnits.METER):
        """
        specifies a series of source location from an inventory object
        :param seeds: a list of locations containing at least the location,
        and location label
        :param units: units of measurement used to express x, y, and z
        :type units: str must be a valid unit of measurement as defined by the
        GridUnits class
        :type seeds: list of dictionary

        :Example:

        >>> seed = Seed('station_code', 'location_code', Coordinates(0, 0, 0))
        >>> seeds = SeedEnsemble([seed])

        """

        self.units = units
        self.seeds = seeds

    def __len__(self):
        return len(self.seeds)

    def __getitem__(self, item):
        return self.seeds[item]

    def __repr__(self):
        return f'{self.units}\n' + '\n'.join(self.seeds)

    def __iter__(self):
        return iter(self.seeds)

    @classmethod
    def from_inventory(cls, inventory: Inventory, units: GridUnits = GridUnits.METER) \
            -> 'SeedEnsemble':
        """
        create from an inventory object
        :param inventory: an inventory object
        :type inventory: uquake.core.inventory.Inventory
        :param units: units of measurement used to express x, y, and z
        :type units: str must be a valid unit of measurement as defined by the
        GridUnits class
        :rparam: a Seeds object
        """

        srces = []
        for instrument in inventory.instruments:
            srce = Seed(instrument.station_code, instrument.location_code,
                        instrument.coordinates)
            srces.append(srce)

        return cls(srces)

    @classmethod
    def from_json(cls, json):
        pass

    def add(self, location: Seed):
        """
        Add a single location to the source list
        :param location:
        :type location: Seed
        """

        self.seeds.append(location)

    @classmethod
    def generate_random_seeds_in_grid(cls, grid, n_seeds=1):
        """
        generate n_seeds random seeds inside the grid provided. This function
        is mainly used for testing purposes
        :param grid: a grid
        :type grid: uquake.grid.base.Grid or an object inheriting from Grid
        :param n_seeds: number of seeds to generate
        :return: a list of seeds

        >>> from uquake.grid.base import Grid
        >>> from uquake.grid.extended import SeedEnsemble
        >>> grid_dimensions = [10, 10, 10]
        >>> grid_spacing = [1, 1, 1]
        >>> grid_origin = [0, 0, 0]
        >>> grid = Grid(grid_dimensions, grid_spacing, grid_origin, value=1)
        >>> seeds = SeedEnsemble.generate_random_seeds_in_grid(grid, n_seeds=10)
        """

        seeds = []
        for seed in Seed.random_in_grid(grid, n_seeds=n_seeds):
            # import ipdb
            # ipdb.set_trace()
            seeds.append(seed)

        return cls(seeds)

    def __repr__(self):
        line = ""

        divider = 1000 if self.units == GridUnits.KILOMETER else 1

        return f'{self.units}\n' + '\n'.join([f'{seed}'
                                              for seed in self.seeds])

    @property
    def nlloc(self):
        line = ""

        divider = 1000 if self.units == GridUnits.KILOMETER else 1

        for seed in self.seeds:
            # test if location name is shorter than 6 characters

            line += f'GTSRCE {seed.label} XYZ ' \
                    f'{seed.x / divider:>15.6f} ' \
                    f'{seed.y / divider:>15.6f} ' \
                    f'{seed.z / divider:>15.6f} ' \
                    f'{seed.elevation:0.2f}\n'

        return line

    @property
    def locs(self):
        seeds = []
        for location in self.locations:
            seeds.append([location['x'], location['y'], location['z']])
        return np.array(seeds)

    @property
    def labels(self):
        seed_labels = []
        for location in self.locations:
            seed_labels.append(location['label'])

        return np.array(seed_labels)


class TypedGrid(Grid):
    """
    base 3D rectilinear grid object
    """

    def __init__(self, data_or_dims, origin, spacing, phase: Phases,
                 value=0, grid_type=GridTypes.VELOCITY_METERS,
                 grid_units=__default_grid_units__,
                 float_type="FLOAT",
                 grid_id: ResourceIdentifier = ResourceIdentifier(),
                 label: str = __default_grid_label__,
                 coordinate_system: CoordinateSystem = CoordinateSystem.NED):
        """
        :param data_or_dims: data or data dimensions. If dimensions are
        provided a homogeneous grid is created with value=value
        :param origin: origin of the grid
        :type origin: list
        :param spacing: the spacing between grid nodes
        :type spacing: list
        :param phase: Phase
        :type phase: Phases
        :param value: a value to fill the grid with if data_or_dims is a list
        :type value: float
        :param grid_type: grid type
        :type grid_type: GridTypes
        :param grid_units: grid units
        :type grid_units: GridUnits
        :param float_type:  the float type either 'FLOAT' or 'DOUBLE'
        :type float_type: FloatTypes
        :param grid_id: the grid id
        :type grid_id: str
        :param label: label of the grid
        :type label: str
        :param coordinate_system: coordinate system either NED or ENU
        :type coordinate_system: CoordinateSystem
        """

        super().__init__(data_or_dims, spacing=spacing, origin=origin,
                         value=value, resource_id=grid_id, label=label,
                         coordinate_system=coordinate_system)

        self.phase = phase
        self.grid_type = grid_type
        self.grid_units = grid_units
        self.float_type = float_type

    @property
    def grid_id(self):
        return self.resource_id

    def mv(self, base_name, origin, destination):
        """
        move a NLLoc grid with a certain base_name from an origin to a
        destination
        :param NLLocGridObject:
        :type NLLocGridObject: uquake.grid.extended.TypedGrid
        :param base_name:
        :type base_name: str
        :param origin:
        :type origin: str
        :param destination:
        :type destination: str
        :return:
        """

        self.write(base_name, destination)
        for ext in self.extensions:
            shutil.move(f'{origin}/{base_name}.{ext}',
                        f'{destination}/{base_name}.{ext}')

    @classmethod
    def from_ods(cls, origin, dimensions, spacing, phase, val=0):
        grid = super().from_ods(origin, dimensions, spacing, val=val)
        return cls(grid.data, origin, spacing, phase)

    @classmethod
    def from_ocs(cls, origin, corner, spacing, phase, val=0):
        grid = super().from_ocs(origin, corner, spacing, val=val)
        return cls(grid.data, grid.origin, grid.spacing, phase)

    @classmethod
    def from_ocd(cls, origin, corner, dimensions, phase, val=0):
        grid = super().from_ocd(origin, corner, dimensions, val=val)
        return cls(grid.data, grid.origin, grid.spacing, phase)

    @property
    def model_id(self):
        return self.resource_id


class Direction(Enum):
    UP = 'UP'
    DOWN = 'DOWN'


class ModelLayer:
    """
    1D model varying in Z
    """

    def __init__(self, z_top, value_top):
        """
        :param z_top: Top of the layer z coordinates
        :param value_top: Value at the top of the layer
        """
        self.z_top = z_top
        self.value_top = value_top

    def __repr__(self):
        return f'top - {self.z_top:5.0f} | value - {self.value_top:5.0f}\n'


class LayeredVelocityModel(object):

    def __init__(self, network_code: str, velocity_model_layers: List = None,
                 phase: Phases = Phases.P,
                 grid_units: GridUnits = __default_grid_units__,
                 float_type=__default_float_type__,
                 gradient=False, grid_id: ResourceIdentifier = ResourceIdentifier(),
                 coordinate_system: CoordinateSystem = CoordinateSystem.NED,
                 label=__default_grid_label__):
        """
        Initialize
        :param network_code: network code
        :type network_code: str
        :param velocity_model_layers: a list of VelocityModelLayer
        :type velocity_model_layers: list
        :param phase: Phase either Phases.P or Phases.S
        :type phase: Phases default = Phases.P
        :param grid_units: units of measurement used to express x, y, and z
        :type grid_units: GridUnits
        :param float_type: float type either 'FLOAT' or 'DOUBLE'
        :type float_type: FloatTypes default = 'FLOAT'
        :param gradient: whether the model is a gradient model or not
        (no control layer by layer)
        :type gradient: bool default = False
        :param grid_id: the grid id
        :type grid_id: uquake.core.event.ResourceIdentifier
        :param coordinate_system: coordinate system either NED or ENU
        :type coordinate_system: CoordinateSystem
        :param label: grid label
        :type label: str
        """

        self.network_code = network_code

        if velocity_model_layers is None:
            self.velocity_model_layers = []

        if isinstance(phase, Phases):
            self.phase = phase
        elif isinstance(phase, str):
            self.phase = Phases(phase)
        else:
            raise TypeError('phase must be a Phases object')

        if isinstance(grid_units, GridUnits):
            self.grid_units = grid_units
        elif isinstance(grid_units, str):
            self.grid_units = GridUnits(grid_units)
        else:
            raise TypeError('grid_units must be a GridUnits object')

        if isinstance(float_type, FloatTypes):
            self.float_type = float_type
        elif isinstance(float_type, str):
            self.float_type = FloatTypes(float_type)
        else:
            raise TypeError('float_type must be a FloatTypes object')

        self.grid_type = GridTypes.VELOCITY_METERS

        self.grid_id = grid_id

        self.gradient = gradient
        self.coordinate_system = coordinate_system
        self.label = label

    def __repr__(self):
        output = ''
        for i, layer in enumerate(self.velocity_model_layers):
            output += f'layer {i + 1:4d} | {layer}'

        return output

    def add_layer(self, layer: ModelLayer):
        """
        Add a layer to the model. The layers must be added in sequence from the
        top to the bottom
        :param layer: a ModelLayer object
        """
        if not (type(layer) is ModelLayer):
            raise TypeError('layer must be a VelocityModelLayer object')

        if self.velocity_model_layers is None:
            self.velocity_model_layers = [layer]
        else:
            self.velocity_model_layers.append(layer)

    def to_1d_model(self, z_min, z_max, spacing):
        # sort the layers to ensure the layers are properly ordered
        z = []
        v = []
        for layer in self.velocity_model_layers:
            z.append(layer.z_top)
            v.append(layer.value_top)

        if np.max(z) < z_max:
            i_z_max = np.argmax(z)
            v_z_max = v[i_z_max]

            z.append(z_max)
            v.append(v_z_max)

        if np.min(z) > z_min:
            i_z_min = np.argmin(z)
            v_z_min = v[i_z_min]

            z.append(z_min)
            v.append(v_z_min)

        i_sort = np.argsort(z)

        z = np.array(z)
        v = np.array(v)

        z = z[i_sort]
        v = v[i_sort]

        z_interp = np.arange(z_min, z_max, spacing)
        kind = 'previous'
        if self.gradient:
            kind = 'linear'

        f_interp = interp1d(z, v, kind=kind)

        v_interp = f_interp(z_interp)

        return z_interp, v_interp

    def to_3d_grid(self, dims, origin, spacing):
        model_grid_3d = VelocityGrid3D.from_layered_model(self, dims, origin,
                                                          spacing, label=self.label)
        return model_grid_3d

    def plot(self, z_min, z_max, spacing, invert_z_axis=True, *args, **kwargs):
        """
        Plot the 1D velocity model
        :param z_min: lower limit of the model
        :param z_max: upper limit of the model
        :param spacing: plotting resolution in z
        :param invert_z_axis: whether the z axis is inverted or not
        (default = True)
        :return: matplotlib axis
        """

        z_interp, v_interp = self.to_1d_model(z_min, z_max, spacing)

        x_label = None
        if self.phase == 'P':
            x_label = 'P-wave velocity'
        elif self.phase == 'S':
            x_label = 's-wave velocity'

        if self.grid_units == 'METER':
            units = 'm'
        else:
            units = 'km'

        y_label = f'z [{units}]'
        ax = plt.axes()
        ax.plot(v_interp, z_interp, *args, **kwargs)
        plt.xlabel(x_label)
        plt.ylabel(y_label)

        if invert_z_axis:
            ax.invert_yaxis()

        ax.set_aspect(2)

        plt.tight_layout()

        return ax


class VelocityGrid3D(TypedGrid):

    def __init__(self, network_code, data_or_dims, origin, spacing,
                 phase: Phases = Phases.P, value=0, float_type=__default_float_type__,
                 grid_id: ResourceIdentifier = ResourceIdentifier(),
                 grid_type=GridTypes.VELOCITY_METERS,
                 grid_units=__default_grid_units__,
                 coordinate_system=CoordinateSystem.NED,
                 label: str = __default_grid_label__, **kwargs):

        """

        :param network_code:
        :param data_or_dims:
        :param origin:
        :param spacing:
        :param phase:
        :param value:
        :param float_type:
        :param grid_id:
        :param grid_type:
        :param grid_units:
        param coordinate_system:
        :param label:
        :param kwargs:
        """

        self.network_code = network_code

        if (type(spacing) is int) | (type(spacing) is float):
            spacing = [spacing, spacing, spacing]

        super().__init__(data_or_dims, origin, spacing, phase,
                         value=value, grid_type=grid_type,
                         grid_units=grid_units, coordinate_system=coordinate_system,
                         float_type=float_type,
                         grid_id=grid_id, label=label)

    @staticmethod
    def get_base_name(network_code, phase):
        """
        return the base name given a network code and a phase
        :param network_code: Code of the network
        :type network_code: str
        :param phase: Phase, either P or S
        :type phase: Phases
        :return: the base name
        """
        return f'{network_code.upper()}.{phase}.mod'

    @classmethod
    def from_ocd(cls, origin, corner, dimensions, val=0):
        pass

    @classmethod
    def from_ocs(cls, origin, corner, spacing, val=0):
        pass

    @classmethod
    def from_ocd(cls, origin, dimensions, spacing, val=0):
        pass

    @classmethod
    def from_layered_model(cls, layered_model, dims, origin,
                           spacing, label=__default_grid_label__, **kwargs):
        """
        Generating a 3D grid model from
        :param layered_model: a LayeredVelocityModel object
        :param dims: dimensions of the grid
        :param origin: origin of the grid
        :param spacing: spacing of the grid
        :param label: label of the grid
        :param kwargs: additional arguments
        :return: a VelocityGrid3D object
        """

        z_min = origin[-1]
        z_max = z_min + spacing[-1] * dims[-1]

        z_interp, v_interp = layered_model.to_1d_model(z_min, z_max, spacing[2])

        data = np.zeros(dims)

        for i, v in enumerate(v_interp):
            data[:, :, i] = v_interp[i]

        return cls(layered_model.network_code, data, origin, spacing,
                   phase=layered_model.phase,
                   float_type=layered_model.float_type,
                   grid_id=layered_model.grid_id, label=label, **kwargs)

    def to_slow_lens(self):
        data = self.spacing[0] / self.data

        return TypedGrid(data, self.origin, self.spacing,
                         self.phase, grid_type='SLOW_LEN',
                         grid_units=self.grid_units,
                         float_type=self.float_type,
                         grid_id=self.grid_id, label=self.label)

    @classmethod
    def from_slow_len(cls, grid: TypedGrid, network_code: str):
        data = np.mean(grid.spacing) / grid.data
        return cls(network_code, data, grid.origin, grid.spacing,
                   phase=grid.phase, float_type=grid.float_type,
                   grid_id=grid.grid_id, label=grid.label)

    def to_time(self, seed: Seed, sub_grid_resolution=0.1,
                *args, **kwargs):
        """
        Eikonal solver based on scikit fast marching solver
        :param seed: numpy array location of the seed or origin of useis wave
         in model coordinates
        (usually location of a station or an event)
        :type seed: numpy.ndarray or list
        :param sub_grid_resolution: resolution of the grid around the seed.
        Propagating the wavefront on a denser grid around the seed,
        significantly improves the travel time accuracy. The value represents
        a fraction of the grid resolution. For instance, assuming a grid with
        spacing of 10m, if the sub_grid_resolution is set to 0.1, the
        resolution around the grid will be 1m.

        :rtype: TTGrid
        """

        if isinstance(seed, list):
            seed = np.array(seed)

        if not self.in_grid(seed.loc):
            logger.warning(f'{seed.label} is outside the grid. '
                           f'The travel time grid will not be calculated')
            return

        origin = self.origin
        shape = self.shape
        spacing = self.spacing

        sub_grid_spacing = spacing * sub_grid_resolution

        # extent = ((4 * sub_grid_spacing) * 1.2 + sub_grid_spacing)

        n_pts_inner_grid = (4 * spacing / sub_grid_spacing * 1.2).astype(int)
        for i in range(0, len(n_pts_inner_grid)):
            if n_pts_inner_grid[i] % 2:
                n_pts_inner_grid[i] += 1

        x_i = np.arange(0, n_pts_inner_grid[0]) * sub_grid_spacing[0]
        y_i = np.arange(0, n_pts_inner_grid[1]) * sub_grid_spacing[1]
        z_i = np.arange(0, n_pts_inner_grid[2]) * sub_grid_spacing[2]

        x_i = x_i - np.mean(x_i) + seed.x
        y_i = y_i - np.mean(y_i) + seed.y
        z_i = z_i - np.mean(z_i) + seed.z

        X_i, Y_i, Z_i = np.meshgrid(x_i, y_i, z_i, indexing='ij')

        coords = np.array([X_i.ravel(), Y_i.ravel(), Z_i.ravel()]).T

        vel = self.interpolate(coords, grid_space=False).reshape(
            X_i.shape)

        phi = np.ones_like(X_i)
        phi[int(np.floor(len(x_i) / 2)), int(np.floor(len(y_i) / 2)),
            int(np.floor(len(z_i) / 2))] = 0

        tt_tmp = skfmm.travel_time(phi, vel, dx=sub_grid_spacing)

        tt_tmp_grid = TTGrid(self.network_code, tt_tmp, [x_i[0], y_i[0], z_i[0]],
                             sub_grid_spacing, seed, self.grid_id,
                             phase=self.phase,
                             float_type=self.float_type, label=self.label)

        # __init__(self, data_or_dims, origin, spacing, seed: Seed,
        # phase: Phases = Phases.P, value: float = 0,
        # float_type: FloatTypes = __default_float_type__,
        # grid_id: ResourceIdentifier = ResourceIdentifier(),
        # grid_units: GridUnits = __default_grid_units__):

        data = self.data

        xe = origin[0] + np.arange(0, shape[0], 1) * spacing[0]
        ye = origin[1] + np.arange(0, shape[1], 1) * spacing[1]
        ze = origin[2] + np.arange(0, shape[2], 1) * spacing[2]

        Xe, Ye, Ze = np.meshgrid(xe, ye, ze, indexing='ij')

        coords = np.array([Xe.ravel(), Ye.ravel(), Ze.ravel()])

        corner1 = np.array([np.min(x_i), np.min(y_i), np.min(z_i)])
        corner2 = np.array([np.max(x_i), np.max(y_i), np.max(z_i)])

        test = ((coords[0, :] >= corner1[0]) & (coords[0, :] <= corner2[0]) &
                (coords[1, :] >= corner1[1]) & (coords[1, :] <= corner2[1]) &
                (coords[2, :] >= corner1[2]) & (coords[2, :] <= corner2[2]))

        Xe_grid = Xe.ravel()[test]
        Ye_grid = Ye.ravel()[test]
        Ze_grid = Ze.ravel()[test]

        X = np.array([Xe_grid, Ye_grid, Ze_grid]).T

        tt_interp = tt_tmp_grid.interpolate(X, grid_space=False,
                                            order=3)[0]

        bias = np.max(tt_interp)

        phi_out = np.ones_like(Xe).ravel()
        phi_out[test] = tt_interp - bias

        phi_out = phi_out.reshape(Xe.shape)

        tt_out = skfmm.travel_time(phi_out, data, dx=spacing)

        # tt_out = tt_out.ravel() + bias
        tt_out = tt_out.ravel() + bias
        tt_out[test] = tt_interp
        tt_out = tt_out.reshape(Xe.shape)

        tt_out_grid = TTGrid(self.network_code, tt_out, self.origin,
                             self.spacing, seed, phase=self.phase,
                             float_type=self.float_type,
                             velocity_model_id=self.grid_id,
                             label=self.label)

        tt_out_grid.data -= tt_out_grid.interpolate(seed.T,
                                                    grid_space=False,
                                                    order=3)[0]

        return tt_out_grid

    def to_time_multi_threaded(self, seeds: SeedEnsemble, cpu_utilisation=0.9,
                               *args, **kwargs):
        """
        Multithreaded version of the Eikonal solver
        based on scikit fast marching solver
        :param seeds: array of seed
        :type seeds: np.ndarray
        :param cpu_utilisation: fraction of the cpu core to be used for the
        processing task (between 0 and 1)
        :type cpu_utilisation: float between 0 and 1
        :param args: arguments to be passed directly to skfmm.travel_time
        function
        :param kwargs: keyword arguments to be passed directly to
        skfmm.travel_time function
        :return: a travel time grid ensemble
        :rtype: TravelTimeEnsemble
        """

        num_threads = int(np.ceil(cpu_utilisation * __cpu_count__))
        # ensuring that the number of threads is comprised between 1 and
        # __cpu_count__
        num_threads = np.max([np.min([num_threads, __cpu_count__]), 1])

        data = []
        for seed in seeds:
            if not self.in_grid(seed.loc):
                logger.warning(f'{seed.label} is outside the grid. '
                               f'The travel time grid will not be calculated')
                continue

        with Pool(num_threads) as pool:
            results = pool.map(self.to_time, seeds)

        tt_grid_ensemble = TravelTimeEnsemble(results)

        return tt_grid_ensemble

    def write(self, path='.'):

        base_name = self.base_name
        super().write(base_name, path=path)

    def mv(self, origin, destination):
        """
        move the velocity grid files from {origin} to {destination}
        :param origin: origin
        :param destination:
        :return:
        """

        super().mv(self, self.base_name,
                   origin, destination)

    @property
    def base_name(self):
        return self.get_base_name(self.network_code, self.phase)


class VelocityGridEnsemble:
    def __init__(self, p_velocity_grid, s_velocity_grid):
        """

        :param p_velocity_grid: p-wave 3D velocity grid
        :type p_velocity_grid: VelocityGrid3D
        :param s_velocity_grid: s-wave 3D velocity grid
        :type s_velocity_grid: VelocityGrid3D

        :NOTE: the p and s velocity grids must have the same dimensions and the same
        label
        """

        if p_velocity_grid.dims != s_velocity_grid.dims:
            raise ValueError('p and s velocity grids must have the same '
                             'dimensions')
        if p_velocity_grid.label != s_velocity_grid.label:
            raise ValueError('p and s velocity grids must have the same '
                             'label')

        self.p_velocity_grid = p_velocity_grid
        self.s_velocity_grid = s_velocity_grid
        self.label = p_velocity_grid.label
        self.__i__ = 0

    def __getitem__(self, item):
        if item.upper() == 'P':
            return self.p_velocity_grid

        elif item.upper() == 'S':
            return self.s_velocity_grid

        else:
            raise ValueError(f'{item} is not a valid key. '
                             f'The key value must either be "P" or "S"')

    def __iter__(self):
        self.__i__ = 0
        return self

    def __next__(self):
        if self.__i__ < 2:
            if self.__i__ == '0':
                return self.p_velocity_grid
            elif self.__i__ == '1':
                return self.s_velocity_grid
        else:
            raise StopIteration

    @staticmethod
    def keys():
        return ['P', 'S']

    def write(self, path='.'):
        for key in self.keys():
            self[key].write(path=path)

    def to_time_multi_threaded(self, seeds: SeedEnsemble, cpu_utilisation=0.9,
                               *args, **kwargs):
        """
        Multithreaded version of the Eikonal solver
        :param seeds:
        :param cpu_utilisation:
        :param args:
        :param kwargs:
        :return:
        """

        tt_grid_ensemble = TravelTimeEnsemble([])

        for key in self.keys():
            tt_grids = self[key].to_time_multi_threaded(seeds,
                                                        cpu_utilisation=
                                                        cpu_utilisation,
                                                        *args, **kwargs)

            tt_grid_ensemble += tt_grids

        return tt_grid_ensemble

    def to_time(self, seeds: SeedEnsemble, multi_threaded=False, *args, **kwargs):
        if multi_threaded:
            return self.to_time_multi_threaded(seeds)

        else:
            tt_grid_ensemble = TravelTimeEnsemble([])
            for key in self.keys():
                for seed in zip(seeds):
                    tt_grid_ensemble += self[key].to_time(seed)

    @property
    def p(self):
        return self['P']

    @property
    def s(self):
        return self['S']

    @property
    def P(self):
        return self['P']

    @property
    def S(self):
        return self['S']


class SeededGridType(Enum):
    TIME = 'TIME'
    ANGLE = 'ANGLE'

    def __str__(self):
        return str(self.value)


class SeededGrid(TypedGrid):
    """
    container for seeded grids (e.g., travel time, azimuth and take off angle)
    """

    __doc__ = f'{TypedGrid.__doc__}\n'

    def __init__(self, network_code, data_or_dims, origin, spacing, seed: Seed,
                 velocity_model_id: ResourceIdentifier,
                 phase: Phases = Phases.P, value: float = 0,
                 grid_units: GridUnits = __default_grid_units__,
                 grid_type: SeededGridType = SeededGridType.TIME,
                 float_type: FloatTypes = __default_float_type__,
                 grid_id: ResourceIdentifier = ResourceIdentifier(),
                 label: str = __default_grid_label__,
                 coordinate_system: CoordinateSystem = CoordinateSystem.NED):
        """
        :param network: network code
        :param data_or_dims: data or data dimensions.
        If dimensions not provided a grid with value=value is created
        :param origin: origin of the grid
        :param spacing: the spacing between grid nodes
        :param seed: seed location of the grid - The seed or instrument location
        :param velocity_model_id: velocity model id
        :param phase: Seismic Phase
        :param value: value of the grid when only specifying the dimensions
        :param grid_units: units of measurement used to express values
        :param grid_type: type of grid (e.g., travel time, azimuth, take off angle)
        :param float_type: float type either 'FLOAT' or 'DOUBLE'
        :param grid_id: the grid id
        :param label: label of the grid
        :param coordinate_system: coordinate system either NED or ENU
        """

        self.network = network_code
        self.seed = seed
        self.velocity_model_id = velocity_model_id

        self.seed_label = seed.label

        if isinstance(grid_type, str):
            self.grid_type = SeededGridType(grid_type)
        elif isinstance(grid_type, SeededGridType):
            self.grid_type = grid_type
        elif isinstance(grid_type, GridTypes):
            self.grid_type = SeededGridType(grid_type.value)
        else:
            raise TypeError('grid_type must be a SeededGridType object')

        super().__init__(data_or_dims, origin, spacing,
                         phase=phase, value=value,
                         grid_type=grid_type, grid_units=grid_units,
                         float_type=float_type, grid_id=grid_id, label=label,
                         coordinate_system=coordinate_system)

        # ensure the data are expressed in the appropriate float_type
        self.data.astype(float_type.value)

    def __repr__(self):
        line = f'{self.grid_type} Grid\n' \
               f'    origin        : {self.origin}\n' \
               f'    spacing       : {self.spacing}\n' \
               f'    dimensions    : {self.shape}\n' \
               f'    seed label    : {self.seed_label}\n' \
               f'    seed location : {self.seed}'
        return line

    @property
    def network_code(self):
        return self.network

    @property
    def station_code(self):
        return self.seed.station_code

    @property
    def location_code(self):
        return self.seed.location_code

    @property
    def instrument_code(self):
        return self.seed.label

    @classmethod
    def get_base_name(cls, network_code, phase, seed_label, grid_type):
        if grid_type not in cls.__valid_grid_type__:
            raise ValueError(f'{grid_type} is not a valid grid type')

        base_name = f'{network_code}.{phase}.{seed_label}.' \
                    f'{grid_type.lower()}'
        return base_name

    @property
    def base_name(self):
        base_name = self.get_base_name(self.network_code, self.phase,
                                       self.seed_label, self.grid_type)
        return base_name

    def write(self, path='.'):
        base_name = self.base_name
        self._write_grid_data(base_name, path=path)
        self._write_grid_header(base_name, path=path, seed=self.seed,
                                seed_label=self.seed_label,
                                seed_units=self.grid_units)
        self._write_grid_model_id(base_name, path=path)


class TTGrid(SeededGrid):
    def __init__(self, network_code, data_or_dims, origin, spacing, seed: Seed,
                 velocity_model_id: ResourceIdentifier,
                 phase: Phases = Phases.P, value: float = 0,
                 float_type: FloatTypes = __default_float_type__,
                 grid_id: ResourceIdentifier = ResourceIdentifier(),
                 label=__default_grid_label__,
                 coordinate_system: CoordinateSystem = CoordinateSystem.NED):

        super().__init__(network_code, data_or_dims, origin, spacing, seed,
                         velocity_model_id=velocity_model_id, phase=phase,
                         value=value, grid_type=GridTypes.TIME,
                         grid_units=GridUnits.SECOND,
                         float_type=float_type,
                         grid_id=grid_id,
                         coordinate_system = coordinate_system,
                         label=label)

    def to_azimuth(self):
        """
        This function calculate the takeoff angle and azimuth for every
        grid point given a travel time grid calculated using an Eikonal solver
        :return: azimuth and takeoff angles grids
        .. Note: The convention for the takeoff angle is that 0 degree is down.
        """

        gds_tmp = np.gradient(self.data)
        gds = [-gd for gd in gds_tmp]

        azimuth = np.arctan2(gds[0], gds[1]) * 180 / np.pi
        # azimuth is zero northwards

        return AngleGrid(self.waveform_id, azimuth, self.origin, self.spacing,
                         'AZIMUTH', self.seed, seed_label=self.seed_label,
                         phase=self.phase, float_type=self.float_type,
                         model_id=ResourceIdentifier(), grid_units=GridUnits.DEGREE)

    def to_takeoff(self):
        gds_tmp = np.gradient(self.data)
        gds = [-gd for gd in gds_tmp]

        hor = np.sqrt(gds[0] ** 2 + gds[1] ** 2)
        takeoff = np.arctan2(hor, -gds[2]) * 180 / np.pi
        # takeoff is zero pointing down
        return AngleGrid(self.network_code, takeoff, self.origin, self.spacing,
                         'TAKEOFF', self.seed, seed_label=self.seed_label,
                         phase=self.phase, float_type=self.float_type,
                         model_id=self.model_id, grid_units=self.grid_units)

    def to_azimuth_point(self, coord, grid_space=False, mode='nearest',
                         order=1, **kwargs):
        """
        calculate the azimuth angle at a particular point on the grid for a
        given seed location
        :param coord: coordinates at which to calculate the takeoff angle
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param mode: interpolation mode
        :param order: interpolation order
        :return: takeoff angle at the location coord
        """

        return self.to_azimuth().interpolate(coord,
                                             grid_space=grid_space,
                                             mode=mode, order=order,
                                             **kwargs)[0]

    def to_takeoff_point(self, coord, grid_space=False, mode='nearest',
                         order=1, **kwargs):
        """
        calculate the takeoff angle at a particular point on the grid for a
        given seed location
        :param coord: coordinates at which to calculate the takeoff angle
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param mode: interpolation mode
        :param order: interpolation order
        :return: takeoff angle at the location coord
        """
        return self.to_takeoff().interpolate(coord,
                                             grid_space=grid_space,
                                             mode=mode, order=order,
                                             **kwargs)[0]

    def ray_tracer(self, start, grid_space=False, max_iter=1000,
                   arrival_id=None):
        """
        This function calculates the ray between a starting point (start) and an
        end point, which should be the seed of the travel_time grid, using the
        gradient descent method.
        :param start: the starting point (usually event location)
        :type start: tuple, list or numpy.array
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param max_iter: maximum number of iteration
        :param arrival_id: id of the arrival associated to the ray if
        applicable
        :rtype: numpy.ndarray
        """

        return ray_tracer(self, start, grid_space=grid_space,
                          max_iter=max_iter, arrival_id=arrival_id,
                          earth_model_id=self.model_id,
                          network=self.network_code)

    @classmethod
    def from_velocity(cls, seed, seed_label, velocity_grid):
        return velocity_grid.to_time(seed, seed_label)

    def write(self, path='.'):
        return super().write(path=path)

    @property
    def location(self):
        return self.seed_label


class TravelTimeEnsemble:
    def __init__(self, travel_time_grids):
        """
        Combine a list of travel time grids together providing meta
        functionality (multithreaded ray tracing, sorting, travel-time
        calculation for a specific location etc.). It is assumed that
        all grids are compatible, i.e., that all the grids have the same
        origin, spacing and dimensions.
        :param travel_time_grids: a list of TTGrid objects

        :NOTE: The travel time grids must all have the same grid labels if not,
        the object will not be created.
        """

        self.travel_time_grids = []
        for travel_time_grid in travel_time_grids:
            if travel_time_grid.label != travel_time_grids[0].label:
                raise ValueError('all travel time grids must have the same '
                                 'label')
            self.travel_time_grids.append(travel_time_grid)

        self.__i__ = 0

        for tt_grid in self.travel_time_grids:
            try:
                assert tt_grid.check_compatibility(travel_time_grids[0])
            except:
                raise AssertionError('grids are not all compatible')

    def __len__(self):
        return len(self.travel_time_grids)

    def __add__(self, other):
        for travel_time_grid in other.travel_time_grids:
            self.travel_time_grids.append(travel_time_grid)

        return TravelTimeEnsemble(self.travel_time_grids)

    def __iter__(self):
        self.__i__ = 0
        return self

    def __next__(self):
        if self.__i__ < len(self):
            result = self.travel_time_grids[self.__i__]
            self.__i__ += 1
            return result
        else:
            raise StopIteration

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.travel_time_grids[item]
        if isinstance(item, str):
            for tt_grid in self.travel_time_grids:
                if tt_grid.seed_label == item:
                    return tt_grid

    def __repr__(self):
        line = f'Number of travel time grids: {len(self)}'
        return line

    @classmethod
    def from_files(cls, path):
        """
        create a travel time ensemble from files located in a directory
        :param path: the base path to the directory containing the travel time
        files.
        :return:
        """
        tt_grids = []
        for fle in Path(path).glob('*time*.hdr'):
            path = fle.parent
            base_name = '.'.join(fle.name.split('.')[:-1])
            fname = str(Path(path) / base_name)
            tt_grid = read_grid(fname, format='NLLOC',
                                float_type=__default_float_type__)
            tt_grids.append(tt_grid)

        return cls(tt_grids)

    def select(self, seed_labels: Optional[List[str]] = None,
               phases: Optional[List[Phases]] = None):
        """
        return a list of grid corresponding to seed_labels.
        :param seed_labels: seed labels of the travel time grids to return
        :param phases: the phase {'P' or 'S'}, both if None.
        :return: a list of travel time grids
        :rtype: TravelTimeEnsemble
        """

        if (seed_labels is None) and (phases is None):
            return self

        tmp = []
        if seed_labels is None:
            seed_labels = np.unique(self.seed_labels)

        if phases is None:
            phases = [Phases.P.value, Phases.S.value]
        else:
            phases = [phase.value if isinstance(phase, Phases) else Phases(phase).value
                      for phase in phases]

        returned_grids = []
        for travel_time_grid in self.travel_time_grids:
            if travel_time_grid.seed_label in seed_labels:
                if travel_time_grid.phase.value in phases:
                    returned_grids.append(travel_time_grid)

        return TravelTimeEnsemble(returned_grids)

    def sort(self, ascending: bool = True):
        """
        sorting the travel time grid by seed_label
        :param ascending: if true the grids are sorted in ascending order
        :param ascending: bool
        :return: sorted travel time grids.
        :rtype: TravelTimeEnsemble
        """

        i = np.sort(self.seed_labels)

        if not ascending:
            i = i[-1::-1]

        sorted_tt_grids = np.array(self.travel_time_grids)[i]

        return TravelTimeEnsemble(sorted_tt_grids)

    def travel_time(self, seed, grid_space: bool = False,
                    seed_labels: Optional[list] = None,
                    phase: Optional[list] = None):
        """
        calculate the travel time at a specific point for a series of location
        ids
        :param seed: travel time seed
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param seed_labels: a list of locations from which to calculate the
        travel time.
        :param phase: a list of phases for which the travel time need to be
        calculated
        :return: a list of dictionary containing the travel time and location id
        """

        if isinstance(seed, list):
            seed = np.array(seed)

        if grid_space:
            seed = self.travel_time_grids[0].transform_from(seed)

        if not self.travel_time_grids[0].in_grid(seed):
            raise ValueError('seed is outside the grid')

        tt_grids = self.select(seed_labels=seed_labels, phase=phase)

        tts = []
        labels = []
        phases = []
        for tt_grid in tt_grids:
            labels.append(tt_grid.seed_label)
            tts.append(tt_grid.interpolate(seed.T,
                       grid_space=False)[0])
            phases.append(tt_grid.phase)

        tts_dict = {}
        for phase in np.unique(phases):
            tts_dict[phase] = {}

        for label, tt, phase in zip(labels, tts, phases):
            tts_dict[phase][label] = tt

        return tts_dict

    def angles(self, seed, grid_space: bool = False,
                seed_labels: Optional[list] = None,
                phase: Optional[list] = None, **kwargs):
        """
        calculate the azimuth at a specific point for a series of location
        ids
        :param seed: travel time seed
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param seed_labels: a list of locations from which to calculate the
        travel time.
        :param phase: a list of phases for which the travel time need to be
        calculated
        :return: a list of dictionary containing the azimuth and location id
        """

        if isinstance(seed, list):
            seed = np.array(seed)

        if grid_space:
            seed = self.travel_time_grids[0].transform_from(seed)

        if not self.travel_time_grids[0].in_grid(seed):
            raise ValueError('seed is outside the grid')

        tt_grids = self.select(seed_labels=seed_labels, phase=phase)

        azimuths = []
        takeoffs = []
        labels = []
        phases = []
        for tt_grid in tt_grids:
            labels.append(tt_grid.seed_label)
            azimuths.append(tt_grid.to_azimuth_point(seed.T,
                                                     grid_space=False,
                                                     **kwargs))
            takeoffs.append(tt_grid.to_takeoff_point(seed.T,
                                                     grid_space=False,
                                                     **kwargs))
            phases.append(tt_grid.phase)

        azimuth_dict = {}
        takeoff_dict = {}
        for phase in np.unique(phases):
            azimuth_dict[phase] = {}
            takeoff_dict[phase] = {}

        for label, azimuth, takeoff, phase in zip(labels, azimuths, takeoffs,
                                                  phases):
            takeoff_dict[phase][label] = takeoff
            azimuth_dict[phase][label] = azimuth

        angle_dict = {}
        angle_dict['takeoff'] = takeoff_dict
        angle_dict['azimuth'] = azimuth_dict

        return angle_dict

    def ray_tracer(self, starts, seed_label=None, multithreading=False,
                   cpu_utilisation=0.9, grid_space=False, max_iter=1000):
        """

        :param starts: origin of the ray, usually the location of the events
        :param seed_label: a list of seed labels
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param multithreading: if True use multithreading
        :param max_iter: maximum number of iteration
        :param cpu_utilisation: fraction of core to use, between 0 and 1.
        The number of core to be use is bound between 1 and the total number of
        cores
        :return: a list of rays
        :rtype: list
        """

        travel_time_grid = self.select(seed_labels=[seed_label])

        kwargs = {'grid_space': grid_space,
                  'max_iter': max_iter}

        if multithreading:

            ray_tracer_func = partial(ray_tracer, **kwargs)

            num_threads = int(np.ceil(cpu_utilisation * __cpu_count__))
            # ensuring that the number of threads is comprised between 1 and
            # __cpu_count__
            num_threads = np.max([np.min([num_threads, __cpu_count__]), 1])

            data = []
            for start in starts:
                data.append((travel_time_grid, start))

            with Pool(num_threads) as pool:
                results = pool.starmap(ray_tracer_func, data)

            for result in results:
                result.network = self.travel_time_grids[0].network_code

        else:
            results = []
            for start in starts:
                results.append(travel_time_grid.ray_tracer(start, **kwargs))

        return results

    @property
    def seeds(self):
        seeds = []
        for seed_label in self.seed_labels:
            seeds.append(self.select(seed_labels=seed_label)[0].seed)

        return np.array(seeds)

    @property
    def label(self):
        return self.travel_time_grids[0].label

    @property
    def seed_labels(self):
        seed_labels = []
        for grid in self.travel_time_grids:
            seed_labels.append(grid.seed_label)
        return np.unique(np.array(seed_labels))

    @property
    def labels(self):
        labels = []
        for grid in self.travel_time_grids:
            labels.append(grid.label)
        return np.unique(np.array(labels))

    @property
    def shape(self):
        return self.travel_time_grids[0].shape

    @property
    def origin(self):
        return self.travel_time_grids[0].origin

    @property
    def spacing(self):
        return self.travel_time_grids[0].spacing


class AngleGrid(SeededGrid):
    def __init__(self, waveform_id: WaveformStreamID, data_or_dims, origin, spacing,
                 angle_type, seed, seed_label=None, phase='P', value=0,
                 float_type="FLOAT", model_id=None, grid_units='degrees'):
        self.angle_type = angle_type
        super().__init__(waveform_id, data_or_dims, origin, spacing, seed,
                         seed_label=seed_label, phase=phase, value=value,
                         grid_type='ANGLE', float_type=float_type,
                         model_id=model_id, grid_units=grid_units)

    def write(self, path='.'):
        pass



# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: core.py
#  Purpose: Expansion of the obspy.core.trace module
#   Author: uquake development team
#    Email: devs@uquake.org
#
# Copyright (C) 2016 uquake development team
# --------------------------------------------------------------------
"""
Expansion of the obspy.core.trace module

:copyright:
    uquake development team (devs@uquake.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
from abc import ABC
import numpy as np
from obspy.core.trace import Trace as ObspyTrace
from obspy.core.trace import Stats as ObspyStats
from obspy import UTCDateTime
from obspy.core.event import WaveformStreamID
from obspy.core.trace import AttribDict
from uquake.core.util.decorators import update_doc

from .event import Pick
from .util import tools


@update_doc
class Stats(ObspyStats, ABC):
    __doc__ = ObspyStats.__doc__.replace('obspy', 'uquake')

    def __init__(self, stats=None, **kwargs):
        super().__init__(**kwargs)

        if stats:
            for item in stats.__dict__.keys():
                self.__dict__[item] = stats.__dict__[item]

    @property
    def site(self):
        return f'{self.station}.{self.location}'

    def __str__(self):
        """
        Return better readable string representation of Stats object.
        """
        priorized_keys = ['network', 'station', 'location', 'channel',
                          'starttime', 'endtime', 'sampling_rate', 'delta',
                          'npts', 'calib']
        return self._pretty_str(priorized_keys)


class Trace(ObspyTrace, ABC):
    def __init__(self, trace=None, **kwargs):
        super(Trace, self).__init__(**kwargs)

        if trace:
            self.stats = Stats(stats=trace.stats)
            self.data = trace.data

        elif 'header' in kwargs.keys():
            self.stats = Stats(stats=kwargs['header'])

        elif 'stats' in kwargs.keys():
            self.stats = Stats(stats=kwargs['stats'])

        else:
            self.stats = Stats()

    @property
    def sr(self):
        return self.stats.sampling_rate

    @property
    def ppv(self):
        return np.max(np.abs(self.data))

    @property
    def ppa(self):
        return None

    def time_to_index(self, time):
        return int((time - self.stats.starttime) * self.sr)

    def index_to_time(self, index):
        return self.stats.starttime + (index / self.sr)

    def times(self):
        sr = self.stats.sampling_rate

        return np.linspace(0, len(self.data) / sr, len(self.data))

    def plot(self, **kwargs):
        from uquake.imaging.waveform import WaveformPlotting
        waveform = WaveformPlotting(stream=self, **kwargs)

        return waveform.plotWaveform()

    def make_pick(self, pick_time, wlen_search,
                  stepsize, snr_wlens, phase_hint=None):

        ipick = self.time_to_index(pick_time)
        sr = self.stats.sampling_rate
        stepsize_samp = int(stepsize * sr)
        snr_wlens_samp = (snr_wlens * sr).astype(int)
        wlen_search_samp = int(wlen_search * sr)

        new_pick, snr = tools.repick_using_snr(self.data, ipick,
                                               wlen_search_samp, stepsize_samp,
                                               snr_wlens_samp)

        waveform_id = WaveformStreamID(channel_code=self.stats.channel,
                                       station_code=self.stats.station)

        pick = Pick(time=self.index_to_time(newpick), waveform_id=waveform_id,
                    phase_hint=phase_hint, evaluation_mode='automatic',
                    evaluation_status='preliminary', method='snr', snr=snr)

        return pick

    def time_within(self, utime, edge_buf=0.0):
        within = True

        if (utime - edge_buf) < self.stats.starttime:
            within = False
        elif (utime + edge_buf) > self.stats.endtime:
            within = False

        return within

    @classmethod
    def create_from_json(cls, trace_json_object):

        trace_json_object['stats']['starttime'] = UTCDateTime(
            trace_json_object['stats']['starttime'])
        trace_json_object['stats']['endtime'] = UTCDateTime(
            trace_json_object['stats']['endtime'])

        trc = cls(header=trace_json_object['stats'], data=np.array(
            trace_json_object['data'], dtype='float32'))

        return trc

    def to_json(self):
        trace_dict = dict()
        trace_dict['stats'] = dict()

        for key in self.stats.keys():
            if isinstance(self.stats[key], UTCDateTime):
                trace_dict['stats'][key] = self.stats[key].isoformat()
            elif isinstance(self.stats[key], AttribDict):
                trace_dict['stats'][key] = self.stats[key].__dict__
            else:
                trace_dict['stats'][key] = self.stats[key]

        trace_dict['data'] = self.data.tolist()

        return trace_dict



# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: <filename>
#  Purpose: <purpose>
#   Author: <author>
#    Email: <email>
#
# Copyright (C) <copyright>
# --------------------------------------------------------------------


import numpy as np
from uuid import uuid4
from ..core.logging import logger
from pkg_resources import load_entry_point
from ..core.util import ENTRY_POINTS
from pathlib import Path
from scipy.ndimage.interpolation import map_coordinates
import matplotlib.pyplot as plt
from scipy.ndimage.filters import gaussian_filter
from uquake.core.coordinates import CoordinateSystem, Coordinates
from typing import Union, List, Tuple
from uquake.core.inventory import Inventory, Network, Station, Channel
import random
from uquake.core.event import ResourceIdentifier
from copy import deepcopy
from hashlib import sha256


def read_grid(filename, format='PICKLE', **kwargs):
    format = format.upper()

    if format not in ENTRY_POINTS['grid'].keys():
        raise TypeError(f'format {format} is currently not supported '
                        f'for Grid objects')

    format_ep = ENTRY_POINTS['grid'][format]
    read_format = load_entry_point(format_ep.dist.key,
                                   f'uquake.io.grid.{format_ep.name}',
                                   'readFormat')

    return read_format(filename, **kwargs)


__default_grid_label__ = 'Default'


class Grid(object):
    """
    Object containing a regular grid
    """

    def __init__(self, data_or_dims: Union[np.ndarray, List, Tuple],
                 spacing: Union[np.ndarray, List, Tuple] = None,
                 origin: Union[np.ndarray, List, Tuple] = None,
                 resource_id: ResourceIdentifier = ResourceIdentifier(),
                 value: float = 0,
                 coordinate_system: CoordinateSystem = CoordinateSystem.NED,
                 label: str = __default_grid_label__):

        """
        can hold both 2 and 3 dimensional grid
        :param data_or_dims: either a numpy array or a Tuple/List with the grid
        dimensions. If grid dimensions are specified, the grid is initialized
        with value
        :param spacing: a set of two or three values containing the grid spacing
        :type spacing: Tuple, List or numpy.ndarray
        :param origin: a set of two or three values origin of the grid
        :type origin: tuple, List or numpy.ndarray
        :param resource_id: unique identifier for the grid, if set to None,
        :param value: value to fill the grid should dims be specified
        :type value:
        uuid4 is used to define a unique identifier.
        :param coordinate_system: Coordinate system
        :type coordinate_system: ~uquake.core.coordinates.CoordinateSystem
        :param label: Label providing additional information on the grid usage
        :type label: str
        """

        data_or_dims = np.array(data_or_dims)

        if data_or_dims.ndim == 1:
            self.data = np.ones(data_or_dims) * value
        else:
            self.data = data_or_dims

        if resource_id is None:
            self.resource_id = str(uuid4())
        else:
            self.resource_id = resource_id

        if origin is None:
            self.origin = np.zeros(len(self.data.shape))
        else:
            origin = np.array(origin)
            if origin.shape[0] == len(self.data.shape):
                self.origin = origin
            else:
                logger.error(f'origin shape should be {len(self.data.shape)}')
                raise ValueError

        if spacing is None:
            self.spacing = np.ones(len(self.data.shape))
        else:
            spacing = np.array(spacing)
            if spacing.shape[0] == len(self.data.shape):
                self.spacing = spacing
            else:
                logger.error(f'spacing shape should be {len(self.data.shape)}')
                raise ValueError

            self.coordinate_system = coordinate_system

        self.id = id
        self.label = label

    def __hash__(self):
        return hash((tuple(self.data.ravel()), tuple(self.spacing),
                     tuple(self.shape), tuple(self.origin)))

    def __eq__(self, other):
        self.hash == other.hash

    def __getitem__(self, item):
        return self.data[item]

    @property
    def hash(self):
        return self.__hash__()

    @classmethod
    def from_ods(cls, origin, dimensions, spacing, val=0):
        """
        create a grid from origin, dimensions and spacing
        :param origin: grid origin
        :type origin: tuple
        :param dimensions: grid dimension
        :type dimensions: tuple
        :param spacing: spacing between the grid nodes
        :type spacing: float
        :param val: constant value with which to fill the grid
        """

        data = np.ones(tuple(dimensions)) * val
        cls.grid = cls(data, spacing=spacing, origin=origin)

    @classmethod
    def from_ocs(cls, origin, corner, spacing, val=0):
        """
        create a grid from origin, corner and spacing
        :param origin: grid origin (e.g., lower left corner for 2D grid)
        :type origin: tuple or list or numpy.array
        :param corner: grid upper (e.g., upper right corner for 2D grid)
        :type corner: tuple or list or numpy.array
        :param spacing: spacing between the grid nodes
        :type spacing: float
        :param val: constant value with which to fill the grid
        :param buf: buffer around the grid in fraction of grid size
        """
        origin2 = origin
        corner2 = corner

        gshape = tuple([int(np.ceil((c - o) / s))
                        for o, c, s in zip(origin2, corner2, spacing)])
        data = np.ones(gshape) * val
        out = cls(data, spacing=spacing, origin=origin)
        out.fill_homogeneous(val)
        return out

    @classmethod
    def from_ocd(cls, origin, corner, dimensions, val=0):
        """
        create a grid from origin, corner and dimensions
        :param origin: grid origin (e.g., lower left corner for 2D grid)
        :param corner: grid upper (e.g., upper right corner for 2D grid)
        :param dimensions: grid dimensions
        :param val: constant value with which to fill the grid
        :return:
        """

        data = np.ones(dimensions) * val
        spacing = (corner - origin) / (dimensions - 1)
        cls(data, spacing, spacing=spacing, origin=origin)
        return cls

    def __repr__(self):
        repr_str = """
        spacing: %s
        origin : %s
        shape  : %s
        """ % (self.spacing, self.origin, self.shape)
        return repr_str

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return np.all((self.shape == other.shape) &
                      (self.spacing == other.spacing) &
                      np.all(self.origin == other.origin))

    def __mul__(self, other):
        if isinstance(other, Grid):
            if self.check_compatibility(self, other):
                mul_data = self.data * other.data
                return Grid(mul_data, spacing=self.spacing,
                            origin=self.origin)
            else:
                raise ValueError

        else:
            raise TypeError

    def __abs__(self):
        return np.abs(self.data)

    def transform_to(self, values):
        """
        transform model space coordinates into grid space coordinates
        :param values: tuple of model space coordinates
        :type values: tuple
        :rtype: tuple
        """
        coords = (values - self.origin) / self.spacing

        return coords

    def transform_to_grid(self, values):
        """
        transform model space coordinates into grid space coordinates
        :param values: tuple of model space coordinates
        :type values: tuple
        :rtype: tuple
        """

        return self.transform_to(values)

    def transform_from(self, values):
        """
        transform grid space coordinates into model space coordinates
        :param values: tuple of grid space coordinates
        :type values: tuple
        :rtype: tuple
        """
        return values * self.spacing + self.origin

    def transform_from_grid(self, values):
        """
        transform grid space coordinates into model space coordinates
        :param values: tuple of grid space coordinates
        :type values: tuple
        :rtype: tuple
        """

        return self.transform_from(values)

    def check_compatibility(self, other):
        """
        check if two grids are compatible, i.e., have the same shape, spacing
        and origin
        """
        return (np.all(self.shape == other.shape) and
                np.all(self.spacing == other.spacing) and
                np.all(self.origin == other.origin))

    def __get_shape__(self):
        """

        return the shape of the object
        """
        return self.data.shape

    shape = property(__get_shape__)

    def copy(self):
        """
        copy the object using copy.deepcopy
        """
        import copy
        cp = copy.deepcopy(self)
        return cp

    def in_grid(self, point):
        """
        Check if a point is inside the grid
        :param point: the point to check in absolute coordinate (model)
        :type point: tuple, list or numpy array
        :returns: True if point is inside the grid
        :rtype: bool
        """
        corner1 = self.origin
        corner2 = self.origin + self.spacing * np.array(self.shape)

        return np.all((point >= corner1) & (point <= corner2))

    def fill_homogeneous(self, value):
        """
        fill the data with a constant value
        :param value: the value with which to fill the array
        """
        self.data.fill(value)

    def fill_random(self, mean, std, smooth_sigma, seed=None):
        """
        fill the model with random number with a mean of "mean".
        :param mean: mean of the random number
        :param std: standard deviation of the grid
        :param sigma: gaussian smoothing parameter
        :param seed: random seed
        """

        np.random.seed(seed)

        self.data = np.random.randn(self.dims[0], self.dims[1], self.dims[2])
        self.smooth(smooth_sigma)
        self.data = self.data * std / np.std(self.data) + mean

    def generate_points(self, pt_spacing=None):
        """
        Generate points within the grid
        """
        # if pt_spacing is None:
        ev_spacing = self.spacing

        dimensions = np.array(self.shape) * self.spacing / ev_spacing

        xe = np.arange(0, dimensions[0]) * ev_spacing[0] + self.origin[0]
        ye = np.arange(0, dimensions[1]) * ev_spacing[1] + self.origin[1]
        ze = np.arange(0, dimensions[2]) * ev_spacing[2] + self.origin[2]

        Xe, Ye, Ze = np.meshgrid(xe, ye, ze)

        Xe = Xe.reshape(np.prod(Xe.shape))
        Ye = Ye.reshape(np.prod(Ye.shape))
        Ze = Ze.reshape(np.prod(Ze.shape))
        data_e = self.data.reshape(np.prod(self.data.shape))
        return Xe, Ye, Ze, data_e

    def flattens(self):
        return self.generate_points()

    def rbf_interpolation_sensitivity(self, location, epsilon,
                                      threshold=0.1):
        """
        calculate the sensitivity of each element given a location
        :param location: location in model space at which the interpolation
        occurs
        :param epsilon: the standard deviation of the gaussian kernel
        :param threshold: threshold relative to the maximum value below which
        the weights are considered 0.
        :rparam: the sensitivity matrix
        """
        x, y, z, v = self.flattens()

        # calculating the distance between the location and every grid points

        dist = np.linalg.norm([x - location[0], y - location[1],
                               z - location[2]], axis=0)

        sensitivity = np.exp(-(dist / epsilon) ** 2)
        sensitivity[sensitivity < np.max(sensitivity) * threshold] = 0
        sensitivity = sensitivity / np.sum(sensitivity)

        return sensitivity

    def generate_random_points_in_grid(self, n_points=1,
                                       grid_space=False,
                                       seed=None):
        """
        Generate a random set of points within the grid
        :param n_points: number of points to generate (default=1)
        :type n_points: int
        :param grid_space: whether the output is expressed in
        grid coordinates (True) or model coordinates (False)
        (default: False)
        :type grid_space: bool
        :param seed: random seed for reproducibility
        :return: an array of triplet
        """

        np.random.seed(seed)

        points = np.random.uniform(0, 1, (n_points, len(self.data.shape)))

        for i in range(n_points):
            points[i] = points[i] * self.dimensions

        if not grid_space:
            return self.transform_from_grid(points)

        return points

    def interpolate(self, coord, grid_space=True, mode='nearest',
                    order=1, **kwargs):
        """
        This function interpolate the values at a given point expressed
        either in grid or absolute coordinates
        :param coord: Coordinate of the point(s) at which to interpolate
        either in grid or absolute coordinates
        :type coord: list, tuple, numpy.array
        :param grid_space: true if the coordinates are expressed in
        grid space (indices can be float) as opposed to model space
        :param mode: {'reflect', 'constant', 'nearest', 'mirror', 'wrap'},
        optional

        The `mode` parameter determines how the input array is extended
        beyond its boundaries. Default is 'constant'. Behavior for each valid
        value is as follows:

        'reflect' (`d c b a | a b c d | d c b a`)
            The input is extended by reflecting about the edge of the last
            pixel.

        'constant' (`k k k k | a b c d | k k k k`)
            The input is extended by filling all values beyond the edge with
            the same constant value, defined by the `cval` parameter.

        'nearest' (`a a a a | a b c d | d d d d`)
            The input is extended by replicating the last pixel.

        'mirror' (`d c b | a b c d | c b a`)
            The input is extended by reflecting about the center of the last
            pixel.

        'wrap' (`a b c d | a b c d | a b c d`)
            The input is extended by wrapping around to the opposite edge.

        :param order: int, optional
            The order of the spline interpolation, default is 3.
            The order has to be in the range 0-5.
        :type order: int

        :type grid_space: bool
        :rtype: numpy.array
        """

        coord = np.array(coord)

        if not grid_space:
            coord = self.transform_to(coord)

        if len(coord.shape) < 2:
            coord = coord[:, np.newaxis]

        try:
            return map_coordinates(self.data, coord, mode=mode, order=order,
                                   **kwargs)
        except Exception as e:
            # logger.warning(e)
            # logger.info('transposing the coordinate array')
            return map_coordinates(self.data, coord.T, mode=mode, order=order,
                                   **kwargs)

    def fill_from_z_gradient(self, vals, zvals):
        data = self.data
        origin = self.origin
        zinds = [int(self.transform_to([origin[0], origin[1], z_])[2]) for z_
                 in zvals]
        # print(zinds, origin)

        data[:, :, zinds[0]:] = vals[0]
        data[:, :, :zinds[-1]] = vals[-1]

        for i in range(len(zinds) - 1):
            # print(i)
            fill = np.linspace(vals[i + 1], vals[i], zinds[i] - zinds[i + 1])
            data[:, :, zinds[i + 1]:zinds[i]] = fill

    def get_grid_point_coordinates(self, mesh_grid=True):
        """
        """
        x = []
        for i, (dimension, spacing) in \
                enumerate(zip(self.data.shape, self.spacing)):
            v = np.arange(0, dimension) * spacing + self.origin[0]
            x.append(v)

        if not mesh_grid:
            return tuple(x)

        if len(x) == 2:
            return tuple(np.meshgrid(x[0], x[1]))

        if len(x) == 3:
            return tuple(np.meshgrid(x[0], x[1], x[2]))

    def write(self, filename, format='PICKLE', **kwargs):
        """
        write the grid to disk
        :param filename: full path to the file to be written
        :type filename: str
        :param format: output file format
        :type format: str
        """
        format = format.upper()

        Path(filename).parent.mkdirs(parent=True, exist_ok=True)

        if format not in ENTRY_POINTS['grid'].keys():
            raise TypeError(f'format {format} is currently not supported '
                            f'for Grid objects')

        format_ep = ENTRY_POINTS['grid'][format]
        write_format = load_entry_point(format_ep.dist.key,
                                        f'uquake.io.grid.{format_ep.name}',
                                        'writeFormat')

        return write_format(self, filename, **kwargs)

    @classmethod
    def read(cls, filename, format='PICKLE'):
        read_grid(filename, format=format)

    def plot_1D(self, x, y, z_resolution, grid_space=False,
                inventory=None, reverse_y=True):
        """

        :param x: x location
        :param y: y location
        :param z_resolution_m: z resolution in grid units
        :param grid_space:
        :return:
        """

        if not grid_space:
            x, y, z = self.transform_from([x, y, 0])

        zs = np.arange(self.origin[2], self.corner[2], z_resolution)

        coords = []
        for z in zs:
            coords.append(np.array([x, y, z]))

        values = self.interpolate(coords, grid_space=grid_space)

        plt.plot(values, zs)
        if reverse_y:
            plt.gca().invert_yaxis()

        if (inventory):
            z_stas = []
            for network in inventory:
                for station in network:
                    loc = station.loc
                    z_stas.append(loc[2])

            plt.plot([np.mean(values)] * len(z_stas), z_stas, 'kv')



            plt.plot()

            plt.plot()
        plt.show()

    def smooth(self,
               sigma: Union[np.ndarray, List[Union[int, float]], Union[int, float]],
               grid_space: bool = False,
               preserve_statistic: bool = True,
               in_place: bool = True) -> 'YourClass':
        """
        Smooth the data using a Gaussian filter.

        :param sigma: Standard deviation for Gaussian filter.
                      Can be a single number or a numpy array.
        :param grid_space: If True, smoothing is performed in grid space.
        :param preserve_statistic: If True, original mean and standard deviation are preserved.
        :param in_place: If True, modifies the object's data in place. Otherwise, returns a new object.
        :return: self if in_place is True, otherwise a new instance of YourClass.
        """

        # Use self's properties or create a copy for a new object.
        obj = self if in_place else deepcopy(
            self)  # You need to import deepcopy from the copy module

        original_mean = np.mean(obj.data)
        original_std = np.std(obj.data)

        if isinstance(sigma, (int, float)):
            sigma = np.array([sigma] * obj.data.ndim)

        if isinstance(sigma, list):
            sigma = np.array(sigma)

        if not grid_space:
            sigma = sigma / obj.spacing

        smoothed_data = gaussian_filter(obj.data, sigma)

        if preserve_statistic:
            smoothed_mean = np.mean(smoothed_data)
            smoothed_std = np.std(smoothed_data)
            obj.data = (smoothed_data - smoothed_mean) * (
                        original_std / smoothed_std) + original_mean
        else:
            obj.data = smoothed_data

        return obj

    @property
    def ndim(self):
        return self.data.ndim

    @property
    def shape(self):
        return list(self.data.shape)

    @property
    def dims(self):
        return self.shape

    @property
    def dimensions(self):
        return self.shape

    @property
    def corner(self):
        return np.array(self.origin) + np.array(self.shape) * \
               np.array(self.spacing)

    @property
    def checksum(self):
        """
        Compute the SHA-256 checksum of a numpy array.
        """
        if self.data.flags['C_CONTIGUOUS']:
            data_bytes = self.data.data
        else:
            contiguous_data = np.ascontiguousarray(self.data)
            data_bytes = contiguous_data.data
        return sha256(data_bytes).hexdigest()

    def generate_random_inventory_in_grid(self, num_station: int = 1,
                                          ratio_uni_tri: float = 3) -> Inventory:
        """
        Generate a random inventory within the grid.

        :param num_station: Number of stations to place.
        :type num_station: int
        :param ratio_uni_tri: Ratio of uniaxial to tri-axial stations.
        :type ratio_uni_tri: float
        :return: Generated inventory.
        :rtype: Inventory
        """

        # Create a Network
        network = Network(code='XX', description='Generated Network')

        # Generate random points
        points = self.generate_random_points_in_grid(n_points=num_station)

        # Calculate number of uniaxial and triaxial stations based on the ratio

        for i, point in enumerate(points):
            # Create coordinates and a Station
            coordinates = Coordinates(point[0], point[1], point[2],
                                      coordinate_system=self.coordinate_system)

            station = Station(code=f"ST{i:02d}", coordinates=coordinates)

            # Determine if this station should be uniaxial or triaxial
            is_uni = random.choices([True, False], [ratio_uni_tri, 1])[0]

            # Create Channel(s)
            if is_uni:
                channel = Channel(code="HHZ", location_code="00",
                                  coordinates=coordinates)
                station.channels.append(channel)
            else:
                for axis in ["Z", "N", "E"]:
                    channel = Channel(code=f"HH{axis}", location_code=f"00",
                                      coordinates=coordinates)
                    station.channels.append(channel)

            # Append Station to Network
            network.stations.append(station)

        # Create an Inventory
        inventory = Inventory(networks=[network])

        return inventory

    def generate_random_catalog_in_grid(self, num_event: int = 1):
        pass


def angles(travel_time_grid):
    """
    This function calculate the take off angle and azimuth for every grid point
    given a travel time grid calculated using an Eikonal solver
    :param travel_time_grid: travel_time grid
    :type travel_time_grid: ~uquake.core.grid.Grid.
    :rparam: azimuth and takeoff angles grids
    .. Note: The convention for the takeoff angle is that 0 degree is down.
    """

    gds_tmp = np.gradient(travel_time_grid.data)
    gds = [-gd for gd in gds_tmp]

    tmp = np.arctan2(gds[0], gds[1])  # azimuth is zero northwards
    azimuth = travel_time_grid.copy()
    azimuth.type = 'ANGLE'
    azimuth.data = tmp

    hor = np.sqrt(gds[0] ** 2 + gds[1] ** 2)
    tmp = np.arctan2(hor, -gds[2])
    # takeoff is zero pointing down
    takeoff = travel_time_grid.copy()
    takeoff.type = 'ANGLE'
    takeoff.data = tmp

    return azimuth, takeoff


def ray_tracer(travel_time_grid, start, grid_space=False, max_iter=1000,
               arrival_id=None, earth_model_id=None,
               network: str=None):
    """
    This function calculates the ray between a starting point (start) and an
    end point, which should be the seed of the travel_time grid, using the
    gradient descent method.
    :param travel_time_grid: a travel time grid
    :type travel_time_grid: TTGrid
    :param start: the starting point (usually event location)
    :type start: tuple, list or numpy.array
    :param grid_space: true if the coordinates are expressed in
    grid space (indices can be fractional) as opposed to model space
    (x, y, z)
    :param max_iter: maximum number of iteration
    :param arrival_id: id of the arrival associated to the ray if
    applicable
    :type arrival_id: uquake.core.event.ResourceIdentifier
    :param earth_model_id: velocity/earth model id.
    :type earth_model_id: uquake.core.event.ResourceIdentifier
    :param network: network information
    :type network: str
    :rtype: numpy.array
    """

    from uquake.core.event import Ray

    interpolation_order = 1

    if grid_space:
        start = np.array(start)
        start = travel_time_grid.transform_from(start)

    origin = travel_time_grid.origin
    spacing = travel_time_grid.spacing
    end = np.array(travel_time_grid.seed)
    start = np.array(start)

    # calculating the gradient in every dimension at every grid points
    gds = [Grid(gd, origin=origin, spacing=spacing)
           for gd in np.gradient(travel_time_grid.data)]

    dist = np.linalg.norm(start - end)
    cloc = start  # initializing cloc "current location" to start
    spacing = np.linalg.norm(spacing)
    gamma = spacing / 2  # gamma is set to half the grid spacing. This
    # should be
    # sufficient. Note that gamma is fixed to reduce
    # processing time.
    nodes = [start]

    iter_number = 0
    while np.all(dist > spacing / 2):
        if iter_number > max_iter:
            break

        # if dist < spacing * 4:
        #     gamma = spacing / 4

        gvect = np.array([gd.interpolate(cloc, grid_space=False,
                                         order=interpolation_order)[0]
                          for gd in gds])

        if np.linalg.norm(gvect) == 0:
            break

        dr = gamma * gvect / (np.linalg.norm(gvect) + 1e-8)

        if np.linalg.norm(dr) < gamma / 2:
            dr = (dr / np.linalg.norm(dr)) * gamma / 2

        cloc = cloc - dr

        nodes.append(cloc)
        dist = np.linalg.norm(cloc - end)

        iter_number += 1

    nodes.append(end)

    tt = travel_time_grid.interpolate(start, grid_space=False,
                                      order=interpolation_order)[0]

    az = travel_time_grid.to_azimuth_point(start, grid_space=False,
                                           order=interpolation_order)
    toa = travel_time_grid.to_takeoff_point(start, grid_space=False,
                                            order=interpolation_order)

    ray = Ray(nodes=nodes, waveform_id=travel_time_grid.waveform_id,
              arrival_id=arrival_id, phase=travel_time_grid.phase,
              azimuth=az, takeoff_angle=toa, travel_time=tt,
              earth_model_id=earth_model_id, network=network)

    return ray



# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ...core.inventory import (Inventory, Network, Station, Instrument, Channel,
                               SystemResponse)
from ...core import UTCDateTime
from datetime import datetime
from typing import TextIO, Union, List, Dict
from pathlib import Path
from pandas import DataFrame
import string
import random


# The naming convention in the sensor file is not consistent, making it hard
# to translate the .sen file into a more
# standard representation

# The below table is not great
sensor_type_lookup_table = {
    'G3': 'geophone',
    'G1': 'geophone',
    'S3': 'geophone',
    'S1': 'geophone'
}

channel_code_lookup_table = {
    'geophone': 'GP'
}


def create_sensor_dictionary(column_name_line: List, sensor_lines: List) \
        -> DataFrame:
    columns = column_name_line.split()
    sensor_dict = {}

    # initializing the dictionary with the header
    columns[0] = columns[0][2:]
    new_columns = []
    for column in columns:
        if column == 'Orientation':
            new_columns.append('Orientation_N')
            new_columns.append('Orientation_E')
            new_columns.append('Orientation_U')
        else:
            new_columns.append(column)
    columns = new_columns
    for column in columns:
        sensor_dict[column] = []

    sensor_dict['location_code'] = []
    sensor_dict['station_code'] = []

    # filling the dictionary list
    for sensor_line in sensor_lines:
        line_elements = sensor_line.split()
        for i, column in enumerate(columns):
            if column == 'Description':
                try:
                    location_code = int(line_elements[i][-3:] - 1)
                    if location_code < 0:
                        location_code = 0
                    sensor_dict['location_code'].append(location_code)
                    sensor_dict['station_code'].append(line_elements[i][:-4])
                except Exception as e:
                    sensor_dict['location_code'].append(0)
                    sensor_dict['station_code'].append(line_elements[i])

            try:
                line_element = float(line_elements[i])
            except Exception as e:
                line_element = line_elements[i]
            sensor_dict[column].append(line_element)

    return DataFrame(sensor_dict)


def create_channels(sensor_df: DataFrame) -> Dict:
    dict_out = {}
    for row in sensor_df.iterrows():
        row = row[1]
        channel_code = f'{channel_code_lookup_table[sensor_type_lookup_table[row["Type"]]]}{row["Cmp"]:0.0f}'
        location_code = f'{row["location_code"]:02d}'

        # Important Notes:
        # Sampling rate is assumed and given a value of 6000, this information
        # should be available
        # start_date is not defined in the sensor file but is a value that is
        # important
        # end_date is not defined in the sensor file but is a value that is
        # important if the sensor has been decommissioned.

        channel = Channel(code=channel_code, location_code=location_code,
                          x=row['Easting'], y=row['Northing'], z=row['Elev'],
                          sample_rate=6000,
                          start_date=UTCDateTime(datetime.fromtimestamp(0)))
        channel.set_orientation([row['Orientation_E'], row['Orientation_N'],
                                 row['Orientation_U']])
        if sensor_type_lookup_table[row['Type']] == 'geophone':
            # assuming for now that the sensor is critically damped

            system_response = SystemResponse()

            ### Need to handle the gain row['Gain']

            system_response.add_geophone(row['Frequency'], row['Sens'])
            system_response.add_digitizer(max_voltage=row['Vmax'],
                                          digitizer_name='Paladin')

            sensitivity = 1
            for stage in system_response.response_stages:
                sensitivity *= stage.stage_gain

            system_response.sensitivity = sensitivity

            channel.response = system_response.response

            # channel.response = geophone_response(row['Frequency'],
            # row['Gain'], sensitivity=row['Sens'])
            # There might be some change required to convert the data in the
            # waveform files that are expressed in ADC counts. The previous
            # and following line may not be accurate.
            channel.calibration_units = 'COUNT'

            # To be rigorous, the sensor, pre-amplifier, digitizer should be
            # added
            # channel.sensor, channel.pre_amplifier, channel.data_logger.

        if row['station_code'] not in dict_out.keys():
            dict_out[row['station_code']] = []

        dict_out[row['station_code']].append(channel)

    return dict_out


def create_station(channel_dict: Dict):
    stations = []
    for key in channel_dict.keys():
        letters = string.ascii_uppercase
        station_code = ''.join(random.choice(letters) for i in range(4))
        stations.append(Station(code=station_code, channels=channel_dict[key],
                                total_number_of_channels=
                                len(channel_dict[key]),
                                selected_number_of_channels=
                                len(channel_dict[key]),
                                start_date=
                                UTCDateTime(datetime.fromtimestamp(0)),
                                alternate_code=key))
    return stations


def read_esg_sensor_file(sen_file: Union[TextIO, str, Path], network: str = "",
                         header_length: int = 38) -> Inventory:
    if not isinstance(sen_file, TextIO):
        sen_file = open(sen_file)

    # Reading the file and removing the header
    previous_site = None

    sensor_file_lines = sen_file.readlines()
    column_name_line = sensor_file_lines[header_length - 2]
    sensor_lines = sensor_file_lines[header_length:]
    sensor_df = create_sensor_dictionary(column_name_line, sensor_lines)

    dict_channel = create_channels(sensor_df)
    stations = create_station(dict_channel)
    network = Network(code=network, stations=stations)
    inventory = Inventory(networks=[network], source='ESG Solutions',
                          sender='ESG Solutions', module='uQuake',
                          module_uri="")
    return inventory


# if __name__ == '__main__':
#     inventory = read_esg_sensor_file("esg_test.sen", network='GB')
#     inventory.write('ESG_test.xml', format='stationxml')



# Copyright (C) 2023, Jean-Philippe Mercier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import json
from obspy.core.util import AttribDict


class CoordinateSystem(Enum):
    """
    Enum class to specify the coordinate system used in microseismic monitoring.
    This coordinate system enforces a right-hand coordinate system.

    Attributes:
    NED -- North, East, Down coordinate system.
    ENU -- East, North, Up coordinate system.
    """

    NED = 1
    ENU = 2

    def __repr__(self):
        return "North, East, Down (NED)" if self.name == "NED" else "East, North, Up (ENU)"

    def __str__(self):
        return str(self.name)


class CoordinateTransformation:
    """
    Class to handle transformations between a custom coordinate systems and latitude
    longitude.

    :param translation: Translation vector as (dx, dy, dz).
    :type translation: tuple[float, float, float]
    :param rotation: Rotation matrix or Euler angles.
    :type rotation: list[list[float]] or tuple[float, float, float]
    :param epsg_code: EPSG code for the target coordinate system.
    :type epsg_code: int
    :param scaling: Optional scaling factor or factors.
    :type scaling: float or tuple[float, float, float]
    :param reference_elevation: Reference elevation for depth conversions.
    :type reference_elevation: float
    """

    def __init__(
        self, translation=0, rotation=0, epsg_code=None, scaling=None, reference_elevation=0.0
    ):
        self.translation = translation
        self.rotation = rotation
        self.epsg_code = epsg_code
        self.scaling = scaling if scaling is not None else 1.0
        self.reference_elevation = reference_elevation

    def __eq__(self, other):
        """
        Overrides the default equality method to compare CoordinateTransformation instances.

        :param other: Another CoordinateTransformation object.
        :type other: CoordinateTransformation
        :return: True if instances are equal, False otherwise.
        :rtype: bool
        """
        if not isinstance(other, CoordinateTransformation):
            return False

        if self.translation != other.translation:
            return False

        if self.rotation != other.rotation:
            return False

        if self.epsg_code != other.epsg_code:
            return False

        if self.scaling != other.scaling:
            return False

        if self.reference_elevation != other.reference_elevation:
            return False

        return True

    def __repr__(self):
        out_str = f"""
                translation: {self.translation}
                   rotation: {self.rotation}
                  epsg_code: {self.epsg_code}
                    scaling: {self.scaling}
        reference elevation: {self.reference_elevation}
        """

        return out_str

    def convert_elevation_depth(self, z, to_depth=True):
        """
        Convert between elevation and depth using a reference elevation or depth and
        scaling.

        :param z: Input elevation or depth.
        :type z: float
        :param to_depth: If True, convert elevation to depth. If False, convert depth to
        elevation.
        :type to_depth: bool
        :return: Converted depth or elevation.
        :rtype: float
        """
        if to_depth:
            return (self.reference_elevation - z) * self.scaling
        else:
            return self.reference_elevation - (z / self.scaling)

    def apply_transformation(self, x, y, z):
        """
        Apply the coordinate transformation to a point.

        :param x: X-coordinate.
        :type x: float
        :param y: Y-coordinate.
        :type y: float
        :param z: Z-coordinate.
        :type z: float
        :return: Transformed (x, y, z).
        :rtype: tuple[float, float, float]
        """
        # Implement your transformation logic here
        pass

    def invert_transformation(self, lat, lon, z):
        """
        Invert the coordinate transformation for a point.

        :param lat: latitude.
        :type lat: float
        :param lon: longitude.
        :type lon: float
        :param z: Z-coordinate.
        :type z: float
        :return: Inverted (x, y, z).
        :rtype: tuple[float, float, float]
        """
        # Implement your inversion logic here
        pass

    def to_json(self):
        return json.dumps(self.__dict__)

    @classmethod
    def from_json(cls, string):
        return cls(**json.loads(string))


class Coordinates:
    def __init__(
        self,
        x: float,
        y: float,
        z: float,
        coordinate_system: CoordinateSystem = CoordinateSystem.NED,
        transformation: CoordinateTransformation = CoordinateTransformation(),
    ):
        """
        :param x: X-coordinate.
        :type x: float
        :param y: Y-coordinate.
        :type y: float
        :param z: Z-coordinate.
        :type z: float
        :param coordinate_system: Coordinate system, either "NED" or "ENU".
        :type coordinate_system: CoordinateSystem
        :param transformation: Coordinate transformation object.
        :type transformation: CoordinateTransformation
        """
        self.x = x
        self.y = y
        self.z = z
        self.coordinate_system = coordinate_system
        self.transformation = transformation

    @property
    def northing(self):
        """
        Get the northing coordinate.

        :return: Northing coordinate based on system.
        :rtype: float
        """
        return self.y if self.coordinate_system == CoordinateSystem.NED else self.x

    @property
    def easting(self):
        """
        Get the easting coordinate.

        :return: Easting coordinate based on system.
        :rtype: float
        """
        return self.x if self.coordinate_system == CoordinateSystem.NED else self.y

    @property
    def up(self):
        """
        Get the up coordinate.

        :return: Up coordinate based on system.
        :rtype: float
        """
        if self.transformation:
            return self.transformation.convert_elevation_depth(self.z, to_depth=False)
        return -self.z if self.coordinate_system == CoordinateSystem.NED else self.z

    @property
    def down(self):
        """
        Get the down coordinate.

        :return: Down coordinate based on system.
        :rtype: float
        """
        if self.transformation:
            return self.transformation.convert_elevation_depth(self.z, to_depth=True)
        return self.z if self.coordinate_system == CoordinateSystem.NED else -self.z

    @property
    def loc(self):
        return [self.x, self.y, self.z]

    def __repr__(self):
        out_str = f"""
                        x: {self.x: 0.2f}
                        y: {self.y: 0.2f}
                        z: {self.z: 0.2f}
        coordinate system: {self.coordinate_system}
           transformation: {self.transformation}
        """
        return out_str

    def __eq__(self, other):
        if not isinstance(other, Coordinates):
            return False
        return (
            True
            if (
                (self.x == other.x)
                & (self.y == other.y)
                & (self.z == other.z)
                & (self.coordinate_system == other.coordinate_system)
                & (self.transformation == other.transformation)
            )
            else False
        )

    def to_json(self):
        out_dict = {}
        for key in self.__dict__.keys():
            # from ipdb import set_trace
            # set_trace()
            if key == "coordinate_system":
                out_dict[key] = str(self.coordinate_system)
            elif key == "transformation":
                if self.transformation is not None:
                    out_dict[key] = self.transformation.to_json()
                else:
                    out_dict[key] = None
            else:
                out_dict[key] = self.__dict__[key]
        return json.dumps(out_dict)

    @classmethod
    def from_json(cls, json_string):
        in_dict = json.loads(json_string)
        coordinate_system = getattr(CoordinateSystem, in_dict["coordinate_system"])

        if in_dict["transformation"] is not None:
            transformation = CoordinateTransformation.from_json(
                in_dict["transformation"])
        else:
            transformation = None

        return cls(
            in_dict["x"],
            in_dict["y"],
            in_dict["z"],
            coordinate_system=coordinate_system,
            transformation=transformation,
        )

    def to_extra_key(self, namespace="mq"):
        return AttribDict({"value": self.to_json(), "namespace": namespace})

    @classmethod
    def from_extra_key(cls, extra_key):
        return cls.from_json(extra_key["value"])
